import itertools
import numpy as np
import spglib as spg
from . import atoms as atoms_module
from ..input_output.logging_tools import logger
from .utilities import ase_atoms_to_spglib_tuple

logger = logger.getChild('relate_structures')


def align_supercell(supercell, prim, symprec=None):
    """Rotate and translate a supercell configuration such that it is aligned
    with the target primitive cell.

    Parameters
    ----------
    sc : ase.Atoms
        supercell configuration
    prim : ase.Atoms
        target primitive configuration
    symprec : float
        precision parameter forwarded to spglib

    Returns
    -------
    tuple(ase.Atoms, numpy.ndarray, numpy.ndarray)
        aligned supercell configuration as well as rotation matrix
        (`3x3` array) and translation vector (`3x1` array) that relate
        the input to the aligned supercell configuration.
    """

    # TODO: Make sure the input is what we expect

    # find rotation and translation
    R, T = relate_structures(supercell, prim, symprec=symprec)

    # Create the aligned system
    aligned_supercell = rotate_atoms(supercell, R)
    aligned_supercell.translate(T)
    aligned_supercell.wrap()
    return aligned_supercell, R, T


def relate_structures(reference, target, symprec=1e-5):
    """Finds rotation and translation operations that align two structures with
    periodic boundary conditions.

    The rotation and translation in Cartesian coordinates will map the
    reference structure onto the target

    Aligning reference with target can be achieved via the transformations::

        R, T = relate_structures(atoms_ref, atoms_target)
        atoms_ref_rotated = rotate_atoms(atoms_ref, R)
        atoms_ref_rotated.translate(T)
        atoms_ref_rotated.wrap()
        atoms_ref_rotated == atoms_target

    Parameters
    ----------
    reference : ase.Atoms
        The reference structure to be mapped
    target : ase.Atoms
        The target structure

    Returns
    -------
    R : numpy.ndarray
        rotation matrix in Cartesian coordinates (`3x3` array)
    T : numpy.ndarray
        translation vector in Cartesian coordinates
    """

    logger.debug('Reference atoms:')
    _debug_log_atoms(reference)

    reference_primitive_cell = get_primitive_cell(reference, symprec=symprec)

    logger.debug('Reference primitive cell')
    _debug_log_atoms(reference_primitive_cell)

    logger.debug('Target atoms:')
    _debug_log_atoms(target)

    target_primitive_cell = get_primitive_cell(target, symprec=symprec)

    logger.debug('Target primitive cell')
    _debug_log_atoms(target_primitive_cell)

    logger.debug('Sane check that primitive cells can match...')
    _assert_structures_match(reference_primitive_cell, target_primitive_cell)

    logger.debug('Finding rotations...')
    rotations = _find_rotations(reference_primitive_cell.cell,
                                target_primitive_cell.cell)

    logger.debug('Finding transformations...')
    for R in rotations:
        rotated_reference_primitive_cell = \
            rotate_atoms(reference_primitive_cell, R)
        T = _find_translation(rotated_reference_primitive_cell,
                              target_primitive_cell)
        if T is not None:
            break
    else:
        raise Exception(('Found no translation!\n'
                         'Reference primitive cell basis:\n'
                         '{}\n'
                         'Target primitive cell basis:\n'
                         '{}')
                        .format(reference_primitive_cell.basis,
                                target_primitive_cell.basis))

    logger.debug(('Found rotation\n'
                  '{}\n'
                  'and translation\n'
                  '{}')
                 .format(R, T))

    return R, T


def is_rotation(R, cell_metric=None):
    """Checks if rotation matrix is orthonormal

    A cell metric can be passed of the rotation matrix is in scaled coordinates

    Parameters
    ----------
    R : numpy.ndarray
        rotation matrix (`3x3` array)
    cell_metric : numpy.ndarray
        cell metric if the rotation is in scaled coordinates
    """
    if not cell_metric:
        cell_metric = np.eye(3)

    V = cell_metric
    V_inv = np.linalg.inv(V)
    lhs = np.linalg.multi_dot([V_inv, R.T, V, V.T, R, V_inv.T])

    return np.allclose(lhs, np.eye(3), atol=1e-4)  # TODO: tol


def _find_rotations(reference_cell_metric, target_cell_metric):
    """ Generates all proper and improper rotations aligning two cell
    metrics. """

    rotations = []
    V1 = reference_cell_metric
    for perm in itertools.permutations([0, 1, 2]):
        # Make sure the improper rotations are included
        for inv in itertools.product([1, -1], repeat=3):
            V2 = np.diag(inv) @ target_cell_metric[perm, :]
            R = np.linalg.solve(V1, V2).T
            # Make sure the rotation is orthonormal
            if is_rotation(R):
                for R_tmp in rotations:
                    if np.allclose(R, R_tmp):  # TODO: tol
                        break
                else:
                    rotations.append(R)

    assert rotations, ('Found no rotations! Reference cell metric:\n'
                       '{}\n'
                       'Target cell metric:\n'
                       '{}').format(reference_cell_metric, target_cell_metric)

    logger.debug('Found {} rotations'.format(len(rotations)))

    return rotations


def _assert_structures_match(ref, prim):
    """ Asserts the structures are compatible with respect to number of atoms,
    atomic numbers and volume.

    TODO: tol
    """

    if len(ref) != len(prim):
        raise ValueError(
            'Number of atoms in reference primitive cell {} does not match '
            'target primitive {}'.format(len(ref), len(prim)))

    if sorted(ref.numbers) != sorted(prim.numbers):
        raise ValueError('Atomic numbers do not match\nReference: {}\nTarget:'
                         ' {}\n'.format(ref.numbers, prim.numbers))

    if not np.isclose(ref.get_volume(), prim.get_volume()):
        raise ValueError(
            'Volume for reference primitive cell {} does not match target '
            'primitive cell {}\n'.format(ref.get_volume(), prim.get_volume()))


def get_primitive_cell(atoms, to_primitive=True, no_idealize=True,
                       symprec=1e-5):
    """ Gets primitive cell from spglib.

    Parameters
    ----------
    atoms : ase.Atoms
        atomic structure
    to_primitive : bool
        passed to spglib
    no_idealize : bool
        passed to spglib
    """
    if not all(atoms.pbc):
        raise ValueError('atoms must have pbc.')
    atoms_as_tuple = ase_atoms_to_spglib_tuple(atoms)
    spg_primitive_cell = spg.standardize_cell(atoms_as_tuple, to_primitive=True,
                                              no_idealize=True,
                                              symprec=symprec)
    primitive_cell = atoms_module.Atoms(cell=spg_primitive_cell[0],
                                        scaled_positions=spg_primitive_cell[1],
                                        numbers=spg_primitive_cell[2],
                                        pbc=True)
    return primitive_cell


def _debug_log_atoms(atoms):
    logger.debug('cell:\n{}'.format(atoms.cell))
    logger.debug('spos:\n{}'.format(atoms.get_scaled_positions()))
    logger.debug('pos:\n{}'.format(atoms.positions))
    logger.debug('numbers:\n{}'.format(atoms.numbers))


def rotate_atoms(atoms, rotation):
    """Rotates the cell and positions of Atoms and returns a copy

    Parameters
    ----------
    atoms : ase.Atoms
        atomic structure
    rotation : numpy.ndarray
        rotation matrix (`3x3` array)
    """
    cell = np.dot(rotation, atoms.cell.T).T
    positions = np.dot(rotation, atoms.positions.T).T
    return atoms_module.Atoms(cell=cell, positions=positions,
                              numbers=atoms.numbers, pbc=atoms.pbc)


def _find_translation(reference, target):
    """Returns the translation between two compatible atomic structures.

    The two structures must describe the same structure when infinitely
    repeated but differ by a translation.

    Parameters
    ----------
    reference : ase.Atoms
    target : ase.Atoms

    Returns
    -------
    numpy.ndarray or None
        translation vector or `None` if structures are incompatible
    """

    atoms = atoms_module.Atoms(cell=target.cell,
                               positions=reference.positions,
                               numbers=reference.numbers,
                               pbc=True)
    atoms.wrap()

    atoms_atom_0 = atoms[0]
    for atom in target:
        if atoms_atom_0.symbol != atom.symbol:
            continue
        T = atom.position - atoms_atom_0.position
        atoms_copy = atoms.copy()
        atoms_copy.positions += T
        if are_nonpaired_configurations_equal(atoms_copy, target):
            return T
    return None


def are_nonpaired_configurations_equal(atoms1, atoms2):
    """ Checks whether two configurations are identical.  To be considered
    equal the structures must have the same cell metric, elemental
    occupation, scaled positions (modulo one), and periodic boundary
    conditions.

    Unlike the ``__eq__`` operator of :class:`ase.Atoms` the order of the
    atoms does not matter.

    Parameters
    ----------
    atoms1 : ase.Atoms
    atoms2 : ase.Atoms

    Returns
    -------
    bool
        True if atoms are equal, False otherwise

    TODO: tol
    """
    n_atoms = len(atoms1)
    if not (np.allclose(atoms1.cell, atoms2.cell, atol=1e-4) and
            n_atoms == len(atoms2) and
            sorted(atoms1.numbers) == sorted(atoms2.numbers) and
            all(atoms1.pbc == atoms2.pbc)):
        return False
    new_cell = (atoms1.cell + atoms2.cell) / 2
    pos = [a.position for a in atoms1] + [a.position for a in atoms2]
    num = [a.number for a in atoms1] + [a.number for a in atoms2]
    s3 = atoms_module.Atoms(cell=new_cell, positions=pos, numbers=num,
                            pbc=True)
    for i in range(n_atoms):
        for j in range(n_atoms, len(s3)):
            d = s3.get_distance(i, j, mic=True)
            if abs(d) < 1e-4:  # TODO: tol
                if s3[i].number != s3[j].number:
                    return False
                break
        else:
            return False
    return True



import numpy as np
import itertools
import sympy
from .tensors import rotate_tensor
from ..input_output.logging_tools import logger
from .utilities import SparseMatrix


logger = logger.getChild('eigensymmetries')


def create_eigentensors(cs):

    for orbit in cs.orbits:
        prototype_cluster = cs._cluster_list[orbit.prototype_index]

        # The first round of eigentensors are created from the atom label
        # permutation condition
        ets = init_ets_from_label_symmetry(prototype_cluster)

        # Loop over the symmetries and apply them iteratively to reduce the
        # number of eigentensors
        for rotation_index, permutation_index in orbit.eigensymmetries:

            # Rotation amtrix in scaled coordinates
            R = cs.rotation_matrices[rotation_index]
            p = cs._permutations[permutation_index]  # permutation vector

            # Skip if the symmetry is the trivial identity operation
            if np.allclose(R, np.eye(3)) and p == tuple(range(orbit.order)):
                continue

            p_inv = np.argsort(p)

            # Create the intermediate constraint matrix and populate it
            M = SparseMatrix.zeros(3**orbit.order, len(ets))
            populate_constraint_matrix(M, ets, R, p_inv)

            nullspace = M.nullspace()

            new_ets = []
            for solution in nullspace:
                # Normalize the solution vector so it is inly integers
                # Done by multiplying with the greatest common multiple of the
                # denominators present
                solution = renormalize_to_integer(solution)
                # From the solution vector and the old eigentensors a new et is
                # created
                new_ets.append(assemble_new_eigentensor(ets, solution))
            ets = new_ets

            # If the force constant is forbidden by symmetry, break early
            if len(ets) == 0:
                break

        orbit.eigentensors = ets


def populate_constraint_matrix(M, ets, R, p_inv):
    for et_index, et in enumerate(ets):
        tmp = rotate_tensor(et.transpose(p_inv), R)
        tmp = tmp - et
        for element_index, v in enumerate(tmp.flat):
            if v != 0:
                M[element_index, et_index] = v


def init_ets_from_label_symmetry(cluster):
    """ This function creates eigentensor which fulfill the atom label
    symmetry. The only information needed is the corresponding cluster
    """
    order = len(cluster)
    assert sorted(cluster) == list(cluster)
    multiplicities = np.unique(cluster, return_counts=True)[1]
    ets = {}
    # Loop over all elements, represented by multi indices
    for multi_index in itertools.product([0, 1, 2], repeat=order):
        # Sort the multi index based on the multiplicities
        # e.g. cluster [1 1 2]
        # -> [z y x] == [y z x]
        sorted_multi_index = []
        j = 0
        for m in multiplicities:
            sorted_multi_index.extend(sorted(multi_index[j:j + m]))
            j += m
        sorted_multi_index = tuple(sorted_multi_index)
        if sorted_multi_index not in ets:
            ets[sorted_multi_index] = np.zeros([3]*order, dtype=np.int64)
        ets[sorted_multi_index][multi_index] = 1
    return list(ets.values())


def assemble_new_eigentensor(eigentensors, solution):
    new_eigentensor = np.zeros(eigentensors[0].shape, dtype=np.int64)
    for parameter, eigentensor in zip(solution, eigentensors):
        new_eigentensor += np.int64(parameter) * eigentensor
    return new_eigentensor


def renormalize_to_integer(vector):
    gcm = 1  # greates common multiple
    for v in vector:
        if v.is_Integer:
            continue
        den = v.q  # Denomonator
        gcd = sympy.gcd(gcm, den)  # greates common divisor
        tmp = den / gcd
        gcm = gcm * tmp
    # Now all elements should be smallest possible integers
    vector = vector * gcm
    return vector



from hiphive.input_output.logging_tools import logger


logger = logger.getChild(__name__)

default_config = {
    'acoustic_sum_rules': True,
    'symprec': 1e-5,
    'length_scale': 0.1,
    'intprec': 1e-12,
    'standardize_prototype': True,
    'shell_tolerance': 1e-5,
    'max_number_constraint_elements': 1e8
    }


class Config(dict):
    def __init__(self, **kwargs):
        super().__init__(**default_config)
        if 'sum_rules' in kwargs:
            logger.warning('kw sum_rules deprecated.'
                           ' Use acoustic_sum_rules instead')
            kwargs['acoustic_sum_rules'] = kwargs['sum_rules']
            del kwargs['sum_rules']
        for k, v in kwargs.items():
            if k not in default_config:
                raise ValueError('{} is not a valid setting'.format(k))
            if not isinstance(v, type(default_config[k])):
                raise TypeError('Value {} of kw {} is not a valid type'
                                .format(v, k))
        super().update(kwargs)



"""
This module introduces the ForceConstantPotential object which acts as the
finalized force constant model.
"""

import pickle
import tarfile
import numpy as np

from collections import Counter
from typing import Any, Dict, List
from .core.atoms import Atoms
from .force_constant_model import ForceConstantModel
from .core.orbits import Orbit
from .core.orbits import OrientationFamily
from .core.tensors import rotation_to_cart_coord, rotate_tensor
from .input_output.read_write_files import (add_items_to_tarfile_pickle,
                                            add_items_to_tarfile_custom,
                                            add_list_to_tarfile_custom,
                                            read_items_pickle,
                                            read_list_custom)
from .input_output.logging_tools import logger


logger = logger.getChild('ForceConstantPotential')


class ForceConstantPotential:
    """ A finalized force constant model. Can produce force constants for any
    structure compatible with the structure for which the model was set up.

    Parameters
    ----------
    cs : ClusterSpace
        The cluster space the model is based upon
    parameters : numpy.ndarray
        The fitted paramteres
    metadata : dict
        metadata dictionary, will be pickled when object is written to file
    """

    def __init__(self, cs, parameters, metadata=None):

        self._prim = cs.primitive_structure.copy()
        self.cluster_list = cs.cluster_list.copy()
        self.atom_list = cs.atom_list.copy()
        self.orbits = []
        self.spacegroup = cs.spacegroup
        self._config = cs._config
        self.cs_summary = cs.summary

        # add metadata
        if metadata is None:
            metadata = dict()
        self._metadata = metadata
        self._add_default_metadata()

        # Extract the eigentensors from the cluster space and use the paramters
        # to construct the finalized force constants
        parameters = cs._map_parameters(parameters)
        p = 0
        for orb in cs.orbits:
            new_orbit = Orbit()
            fc = np.zeros(orb.eigentensors[0].shape)
            for et, a in zip(orb.eigentensors, parameters[p:]):
                fc += et * a
            new_orbit.force_constant = fc
            new_orbit.order = orb.order
            new_orbit.radius = orb.radius
            new_orbit.maximum_distance = orb.maximum_distance
            for of in orb.orientation_families:
                new_of = OrientationFamily()
                new_of.cluster_indices = of.cluster_indices.copy()
                sym_ind = of.symmetry_index
                R = rotation_to_cart_coord(cs.rotation_matrices[sym_ind],
                                           self.primitive_structure.cell)
                fc = rotate_tensor(new_orbit.force_constant, R.T)
                perm = cs.permutations[of.permutation_indices[0]]
                new_of.force_constant = fc.transpose(perm)
                new_orbit.orientation_families.append(new_of)
            self.orbits.append(new_orbit)
            p += len(orb.eigentensors)

    @property
    def symprec(self):
        return self._config['symprec']

    def write(self, filename):
        """ Writes a ForceConstantPotential to file.

        Parameters
        ----------
        filename : str
            name of file to write ForceConstantPotential to
        """

        # Create a tar archive
        if isinstance(filename, str):
            tar_file = tarfile.open(name=filename, mode='w')
        else:
            raise ValueError('filename must be str')

        # objects with custom write
        add_list_to_tarfile_custom(tar_file, self.orbits, 'orbits')

        # prim with its builtin write/read functions. Note prim is a hiphive.core.Atoms object
        items_custom = {'_prim': self._prim}
        add_items_to_tarfile_custom(tar_file, items_custom)

        # Attributes in pickle format
        pickle_attributes = ['_config', '_metadata', 'spacegroup', 'atom_list',
                             'cluster_list', 'cs_summary']
        items_pickle = dict()
        for attribute in pickle_attributes:
            items_pickle[attribute] = self.__getattribute__(attribute)
        add_items_to_tarfile_pickle(tar_file, items_pickle, 'attributes')

        # Done!
        tar_file.close()

    @staticmethod
    def read(filename):
        """ Reads a ForceConstantPotentialfrom file.

        Parameters
        ----------
        filename : str
            name of input file to load ForceConstantPotential from

        Returns
        -------
        ForceConstantPotential
            the original object as stored in the file
        """

        # Try usage of old read format, Remove in hiphive 1.0
        try:
            old_fcp = ForceConstantPotential._read_old(filename)
            logger.warning('This fcp was written with a version <1.0. Please rewrite it.')
            return old_fcp
        except Exception:
            pass

        # Instantiate empty cs obj.
        fcp = ForceConstantPotential.__new__(ForceConstantPotential)

        # Load from file on disk
        if type(filename) is str:
            tar_file = tarfile.open(mode='r', name=filename)
        else:
            raise ValueError('filename must be str')

        # Attributes with custom read
        fileobj = tar_file.extractfile('_prim')
        fcp._prim = Atoms.read(fileobj)

        fcp.orbits = read_list_custom(tar_file, 'orbits', Orbit.read)

        # Attributes
        attributes = read_items_pickle(tar_file, 'attributes')
        for name, value in attributes.items():
            fcp.__setattr__(name, value)

        # Done!
        tar_file.close()
        return fcp

    # TODO: Remove?
    def _write_old(self, f):
        """Writes a force constant potential to file using old format.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        if isinstance(f, str):
            with open(f, 'wb') as fobj:
                pickle.dump(self, fobj)
        else:
            try:
                pickle.dump(self, f)
            except Exception:
                raise Exception('Failed writing to file.')

    @staticmethod
    def _read_old(f):
        """Reads a force constant potential from file in old format.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        ForceConstantPotential
            the original object as stored in the file
        """
        if isinstance(f, str):
            with open(f, 'rb') as fobj:

                # This allows for reading FCPs with ASE-3.17 and 3.18
                fcp = pickle.load(fobj)
                _prim = fcp._prim

                if hasattr(_prim, '_cell'):  # 3.17
                    cell = _prim._cell
                else:                       # 3.18
                    cell = _prim.cell[:]

                # assume PBC True (as it has to be True in hiphive)
                pbc = [True, True, True]

                new_prim = Atoms(
                    symbols=_prim.symbols, positions=_prim.positions, cell=cell, pbc=pbc)
                fcp._prim = new_prim
                return fcp
        else:
            return pickle.load(f)

    @property
    def metadata(self):
        """ dict : metadata associated with force constant potential """
        return self._metadata

    @property
    def primitive_structure(self):
        """ ase.Atoms : atomic structure """
        return self._prim.copy()

    @property
    def orbit_data(self) -> List[Dict[str, Any]]:
        """list of dictionaries containing detailed information for each
        orbit, e.g. cluster radius and force constant
        """
        data = []
        for orbit_index, orbit in enumerate(self.orbits):
            d = {}
            d['index'] = orbit_index
            d['order'] = orbit.order
            d['radius'] = orbit.radius
            d['maximum_distance'] = orbit.maximum_distance
            d['n_clusters'] = len(orbit.orientation_families)

            types = []
            for atom_ind in self.cluster_list[orbit.prototype_index]:
                types.append(self.primitive_structure.numbers[
                    self.atom_list[atom_ind].site])
            d['prototype_cluster'] = self.cluster_list[orbit.prototype_index]
            d['prototype_atom_types'] = types

            d['geometrical_order'] = len(set(d['prototype_cluster']))
            d['force_constant'] = orbit.force_constant
            d['force_constant_norm'] = np.linalg.norm(orbit.force_constant)
            data.append(d)
        return data

    def print_tables(self):
        """ Prints information concerning the underlying cluster space to stdout,including,
        e.g., the number of cluster, orbits, and parameters by order and number of bodies. """
        self.cs_summary.print_tables()

    def get_force_constants(self, atoms):
        """ Return the force constants of a compatible structure.

        Parameters
        ----------
        atoms : ase.Atoms
            input structure

        Returns
        -------
        ForceConstants
            force constants
        """
        return ForceConstantModel(atoms, self).get_force_constants()

    def __str__(self):
        orbits = self.orbit_data
        orbit_counts = Counter([orbit['order'] for orbit in orbits])
        cluster_counts = Counter()
        for orbit in orbits:
            cluster_counts[orbit['order']] += orbit['n_clusters']

        n = 54
        s = []
        s.append(' ForceConstantPotential '.center(n, '='))
        s.append(f'Spacegroup {self.spacegroup}')
        s.append(f'Cell:\n{self.primitive_structure.cell}')
        s.append(f'Basis:\n{self.primitive_structure.basis}')
        s.append(f'Numbers: {self.primitive_structure.numbers}')

        s.append(f'Cutoff matrix:\n{self.cs_summary._cutoff_matrix}')

        for order in sorted(orbit_counts.keys()):
            n_dofs = self.cs_summary.ndofs_by_order[order]
            s.append(f'Order {order}, #orbits {orbit_counts[order]}, #cluster '
                     f'{cluster_counts[order]}, #parameters {n_dofs}')
        s.append(f'Total number of orbits: {len(orbits)}')
        s.append(f'total number of clusters: {sum(cluster_counts.values())}')
        s.append(f'total number of parameters: {sum(self.cs_summary.ndofs_by_order.values())}')
        s.append(''.center(n, '='))
        return '\n'.join(s)

    def __repr__(self):
        return 'ForceConstantPotential(ClusterSpace({!r}, ...), [...])'.format(
            self.primitive_structure)

    def _add_default_metadata(self):
        """Adds default metadata to metadata dict."""
        import getpass
        import socket
        from datetime import datetime
        from . import __version__ as hiphive_version

        self._metadata['date_created'] = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self._metadata['username'] = getpass.getuser()
        self._metadata['hostname'] = socket.gethostname()
        self._metadata['hiphive_version'] = hiphive_version



import numpy as np
from itertools import permutations, product


def write_fcs_gpumd(fname_fc: str,
                    fname_clusters: str,
                    fcs,
                    order: int,
                    tol: float = 1e-10):
    """
    Writes force constants of given order in GPUMD format.

    Parameters
    ----------
    fname_fc
        name of file which contains the lookup force constants
    fname_clusters
        name of file which contains the clusters and the force constant lookup index
    fcs
        force constants
    order
        force constants for this order will be written to file
    tol
        if the norm of a force constant term is less than this value it will be excluded
        from the output;
        if two force-constants differ by this value or less, they are considered equal.
    """
    cluster_lookup, fc_lookup = _get_lookup_data_smart(fcs, order, tol)
    _write_clusters(fname_clusters, cluster_lookup, order)
    _write_fc_lookup(fname_fc, fc_lookup, order)


def _write_fc_lookup(fname, fc_lookup, order):
    """ Writes the lookup force constants to file """
    fmt = '{}' + ' {}'*order
    with open(fname, 'w') as f:
        f.write(str(len(fc_lookup)) + '\n\n')
        for fc in fc_lookup:
            for xyz in product(range(3), repeat=order):
                f.write(fmt.format(*xyz, fc[xyz])+'\n')
            f.write('\n')


def _write_clusters(fname, cluster_lookup, order):
    """ Writes the cluster lookup to file """
    fmt = '{}' + ' {}'*order
    with open(fname, 'w') as f:
        f.write(str(len(cluster_lookup)) + '\n\n')
        for c, i in cluster_lookup.items():
            line = fmt.format(*c, i) + '\n'
            f.write(line)


def _get_clusters(fcs,
                  order: int,
                  tol: float):
    """ Collect all relevant clusters; for 2nd and 3rd-order force constants
    all permutations are included.
    """
    if order in [2, 3]:
        clusters = []
        for c in fcs._fc_dict.keys():
            if len(c) == order and np.linalg.norm(fcs[c]) > tol:
                for ci in permutations(c):
                    clusters.append(ci)
        clusters = list(sorted(set(clusters)))
    else:
        clusters = [c for c in fcs._fc_dict.keys() if len(c) == order and np.linalg.norm(fcs[c]) > tol]  # noqa
    return clusters


def _get_lookup_data_naive(fcs,
                           order: int,
                           tol: float):
    """ Groups force constants for a given order into groups for which the
    force constant is identical. """
    fc_lookup = []
    cluster_lookup = dict()

    clusters = _get_clusters(fcs, order, tol)

    for c in clusters:
        fc1 = fcs[c]
        if np.linalg.norm(fc1) < tol:
            continue
        for i, fc2 in enumerate(fc_lookup):
            if np.linalg.norm(fc1 - fc2) < tol:
                cluster_lookup[c] = i
                break
        else:
            cluster_lookup[c] = len(fc_lookup)
            fc_lookup.append(fc1)
    return cluster_lookup, fc_lookup


def _get_lookup_data_smart(fcs,
                           order: int,
                           tol: float):
    """ Groups force constants for a given order into groups for which the
    force constant is identical. """
    fc_lookup = []
    cluster_lookup = dict()
    axis = tuple(range(1, order+1))

    clusters = _get_clusters(fcs, order, tol)
    fc_all = np.array([fcs[c] for c in clusters])

    indices = list(range(len(clusters)))
    while len(indices) > 0:
        i = indices[0]
        delta = fc_all[indices] - fc_all[i]
        delta_norm = np.sqrt(np.sum(delta**2, axis=axis))

        inds_to_del = [indices[x] for x in np.where(delta_norm < tol)[0]]
        assert i in inds_to_del

        fc_lookup.append(fc_all[i])
        for j in inds_to_del:
            indices.remove(j)
            cluster_lookup[clusters[j]] = len(fc_lookup)-1
    return cluster_lookup, fc_lookup


def write_fcp_txt(fname: str,
                  path: str,
                  n_types: int,
                  max_order: int,
                  heat_current_order: int = 2):
    """ Write driver potential file for GPUMD.

    Parameters
    ----------
    fname
        file name
    path
        path to directory with force constant file
    n_types
        number of atom types
    max_order
        maximum order of the force constant potential
    heat_current_order
        heat current order used in thermal conductivity


    Format is a simple file containing the following

    fcp number_of_atom_types
    highest_force_order heat_current_order
    path_to_force_constant_files

    which in practice for a binary system with a sixth order model,
    consider third-order heat-currents, would mean

    fcp 2
    6 3
    /path/to/your/folder
    """

    with open(fname, 'w') as f:
        f.write('fcp {}\n'.format(n_types))
        f.write('{} {}\n'.format(max_order, heat_current_order))
        f.write('{}'.format(path.rstrip('/')))  # without a trailing '/'


def write_r0(fname, atoms):
    """
    Write GPUMD r0 file, with reference atomic positions.

    Parameters
    ----------
    fname : str
        name of file to which to write the atomic positions
    atoms : ase.Atoms
        input structure

    """
    line = '{} {} {}\n'
    with open(fname, 'w') as f:
        for a in atoms:
            f.write(line.format(*a.position))



"""
Helper functions for reading and writing objects to tar files
"""

import pickle
import tempfile
import warnings
from tarfile import TarFile
import ase.io as aseIO
from ase import Atoms
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=FutureWarning)
    import h5py


def add_ase_atoms_to_tarfile(tar_file: TarFile,
                             atoms: Atoms,
                             arcname: str,
                             format: str = 'json'):
    """ Adds an ase.Atoms object to tar_file.

    Parameters
    ----------
    tar_file
        tar file to write to
    atoms
        atomic configuration
    arcname
        name of field in archive
    format
        format to use for writing (see ase)
    """
    temp_file = tempfile.NamedTemporaryFile()
    aseIO.write(temp_file.name, atoms, format=format)
    temp_file.seek(0)
    tar_info = tar_file.gettarinfo(arcname=arcname, fileobj=temp_file)
    tar_file.addfile(tar_info, temp_file)


def read_ase_atoms_from_tarfile(tar_file: TarFile,
                                arcname: str,
                                format: str = 'json') -> Atoms:
    """ Reads ase.Atoms from tar file.

    Parameters
    ----------
    tar_file
        tar file from which to read
    arcname
        name of field in archive
    format
        format to use for writing (see ase)
    """
    temp_file = tempfile.NamedTemporaryFile()
    temp_file.write(tar_file.extractfile(arcname).read())
    temp_file.seek(0)
    atoms = aseIO.read(temp_file.name, format=format)
    return atoms


def add_items_to_tarfile_hdf5(tar_file, items, arcname):
    """ Add items to one hdf5 file """
    temp_file = tempfile.NamedTemporaryFile()
    hf = h5py.File(temp_file.name, 'w')
    for key, value in items.items():
        hf.create_dataset(key, data=value, compression='gzip')
    hf.close()
    temp_file.seek(0)
    tar_info = tar_file.gettarinfo(arcname=arcname, fileobj=temp_file)
    tar_file.addfile(tar_info, temp_file)
    temp_file.close()


def add_items_to_tarfile_pickle(tar_file, items, arcname):
    """ Add items by pickling them """
    temp_file = tempfile.TemporaryFile()
    pickle.dump(items, temp_file)
    temp_file.seek(0)
    tar_info = tar_file.gettarinfo(arcname=arcname, fileobj=temp_file)
    tar_file.addfile(tar_info, temp_file)
    temp_file.close()


def add_items_to_tarfile_custom(tar_file, items):
    """ Add items assuming they have a custom write function """
    for key, value in items.items():
        temp_file = tempfile.TemporaryFile()
        value.write(temp_file)
        temp_file.seek(0)
        tar_info = tar_file.gettarinfo(arcname=key, fileobj=temp_file)
        tar_file.addfile(tar_info, temp_file)
        temp_file.close()


def add_list_to_tarfile_custom(tar_file, objects, arcname):
    """ Add list of objects assuming they have a custom write function """
    for i, obj in enumerate(objects):
        temp_file = tempfile.TemporaryFile()
        obj.write(temp_file)
        temp_file.seek(0)
        fname = '{}_{}'.format(arcname, i)
        tar_info = tar_file.gettarinfo(arcname=fname, fileobj=temp_file)
        tar_file.addfile(tar_info, temp_file)
        temp_file.close()


def read_items_hdf5(tar_file, arcname):
    """ Read items from hdf5file inside tar_file """

    # read hdf5
    temp_file = tempfile.NamedTemporaryFile()
    temp_file.write(tar_file.extractfile(arcname).read())
    temp_file.seek(0)
    hf = h5py.File(temp_file.name, 'r')
    items = {key: value[:] for key, value in hf.items()}
    hf.close()
    return items


def read_items_pickle(tar_file, arcname):
    items = dict()
    items = pickle.load(tar_file.extractfile(arcname))
    return items


def read_list_custom(tar_file, arcname, read_function, **kwargs):
    objects = []
    i = 0
    while True:
        try:
            fname = '{}_{}'.format(arcname, i)
            f = tar_file.extractfile(fname)
            obj = read_function(f, **kwargs)
            objects.append(obj)
            f.close()
        except KeyError:
            break
        i += 1
    return objects



# Contains the get_clusters function which generates clusters

import numpy as np
import itertools
from collections import defaultdict

from .utilities import BiMap
from ..input_output.logging_tools import logger

logger = logger.getChild('get_clusters')


# TODO: This function could be made a bit more general
def get_clusters(atoms, cutoffs, nPrim, multiplicity=True,
                 use_geometrical_order=False):
    """Generate a list of all clusters in the atoms object which includes the
    center atoms with positions within the cell metric. The cutoff determines
    up to which order and range clusters should be generated.

    With multiplicity set to True clusters like `[0,0]` and `[3,3,4` etc will
    be generated. This is useful when doing force constants but not so much for
    cluster expansions.

    The geometrical order is the total number of different atoms in the
    cluster. `[0,0,1]` would have geometrical order 2 and `[1,2,3,4]` would
    have order 4. If the key word is True the cutoff criteria will be based on
    the geometrical order of the cluster. This is based on the observation that
    many body interactions decrease fast with cutoff but anharmonic
    interactions can be quite long ranged.

    Parameters
    ----------
    atoms : ase.Atoms
        can be a general atoms object but must have pbc=False.
    cutoffs : dict
        the keys specify the order while the values specify the cutoff radii
    multiplicity : bool
        includes clusters where same atom appears more than once
    geometrical_order : bool
        specifies if the geometrical order should be used as cutoff_order,
        otherwise the normal order of the cluster is used

    Returns
    -------
    list(tuple(int))
        a list of clusters where each entry is a tuple of indices,
        which refer to the atoms in the input supercell
    """

    logger.debug('Generating clusters...')
    cluster_dict = defaultdict(list)
    # Generate all on-site clusters of all orders (1-body)
    for i in range(nPrim):
        for order in cutoffs.orders:
            cluster = (i,) * order
            cluster_dict[order].append(cluster)

    # Generate all 2-body clusters and above in order
    for nbody in cutoffs.nbodies:
        cutoff = cutoffs.max_nbody_cutoff(nbody)
        # Generate all n-body, order n clusters compatible with the cutoff
        nbody_clusters, nbody_cutoffs = generate_geometrical_clusters(atoms, nPrim, cutoff, nbody)
        for order in range(nbody, cutoffs.max_nbody_order(nbody) + 1):
            for cluster, cutoff in zip(nbody_clusters, nbody_cutoffs):
                # If the cutoff of the n-body cluster is compatible with order (order > n) then
                # extend the n-body cluster to higher order (e.g. nbody=3, order=6: ijk -> iijkkk)
                if cutoff < cutoffs.get_cutoff(nbody=nbody, order=order):
                    cluster_dict[order].extend(extend_cluster(cluster, order))

    # The clusters are saved in a BiMap structure which allows for fast lookups
    cluster_list = BiMap()
    for key in sorted(cluster_dict):
        # For each order the clusters are saved in lexicographical order
        for cluster in sorted(cluster_dict[key]):
            cluster_list.append(cluster)
    return cluster_list


def generate_geometrical_clusters(atoms, n_prim, cutoff, order):
    neighbor_matrix, distance_matrix = create_neighbor_matrices(atoms, cutoff)
    clusters, cutoffs = [], []
    i, j = 0, 0
    # The clusters are generated in lexicographical order
    for cluster in itertools.combinations(range(len(atoms)), r=order):
        # If the first atom in the cluster has an index higher or equal to the number of atoms in
        # the primitive cell then no upcoming cluster will have an atom in the primitive cell, thus
        # we can break
        if cluster[0] >= n_prim:
            break
        # if the last cluster failed on index i, j we start by checking this index again to speed
        # things up
        if not neighbor_matrix[cluster[i], cluster[j]]:
            continue
        # loop through all pairs in the cluster and check so that they are neighbors
        for i, j in itertools.combinations(range(order), r=2):
            if not neighbor_matrix[cluster[i], cluster[j]]:
                break
        else:
            clusters.append(cluster)
            # We also note the cutoff each cluster is compatible with
            cutoffs.append(np.max(distance_matrix[cluster, :][:, cluster]))
    return clusters, cutoffs


def create_neighbor_matrices(atoms, cutoff):
    distance_matrix = atoms.get_all_distances(mic=False)  # or True?
    neighbor_matrix = distance_matrix < cutoff
    return neighbor_matrix, distance_matrix


def extend_cluster(cluster, order):
    clusters = []
    cluster = tuple(cluster)
    nbody = len(cluster)
    r = order - nbody
    for tup in itertools.combinations_with_replacement(cluster, r):
        new_cluster = sorted(cluster + tup)
        clusters.append(tuple(new_cluster))
    return clusters



from ase.calculators.calculator import Calculator, all_changes
from ase.neighborlist import NeighborList
import numpy as np

import numba
from numba.typed import List


alpha = 1 / 137.035999046  # no unit
hbar = 6.582119569e-16  # eV * s
c = 299792458 * 1e10  # m/s * A/m
prefactor = alpha * hbar * c  # eV * A


@numba.njit
def a(Z_i, Z_j):
    return 0.46850 / (Z_i**0.23 + Z_j**0.23)


@numba.njit
def phi(x):
    return (0.18175 * np.exp(-3.19980 * x)
            + 0.50986 * np.exp(-0.94229 * x)
            + 0.28022 * np.exp(-0.40290 * x)
            + 0.02817 * np.exp(-0.20162 * x))


@numba.njit
def phi_derivative(x):
    return (- 3.19980 * 0.18175 * np.exp(-3.19980 * x)
            - 0.94229 * 0.50986 * np.exp(-0.94229 * x)
            - 0.40290 * 0.28022 * np.exp(-0.40290 * x)
            - 0.20162 * 0.02817 * np.exp(-0.20162 * x))


@numba.njit
def zbl_energy(Z_i, Z_j, r_ij):
    return prefactor * Z_i * Z_j / r_ij * phi(r_ij / a(Z_i, Z_j))


@numba.njit
def zbl_force(Z_i, Z_j, r_ij):
    return (-prefactor
            * (Z_i * Z_j)
            * (phi_derivative(r_ij / a(Z_i, Z_j)) / (r_ij * a(Z_i, Z_j))
               - phi(r_ij / a(Z_i, Z_j)) / r_ij**2))


@numba.njit
def np_linalg_norm_axis1(mat):
    norm = np.empty(len(mat), dtype=np.float64)
    for i in range(len(mat)):
        norm[i] = np.linalg.norm(mat[i])
    return norm


@numba.njit
def inner_loop(ai, positions, cell, offsets, neighbors, forces, energy, numbers):
    cells = np.dot(offsets, cell)
    v_ij = positions[neighbors] + cells - positions[ai]
    r_ij = np_linalg_norm_axis1(v_ij)
    for aj, v, r in zip(neighbors, v_ij, r_ij):
        energy[0] += zbl_energy(numbers[ai], numbers[aj], r)
        force_j = v / r * zbl_force(numbers[ai], numbers[aj], r)
        forces[aj] += force_j
        forces[ai] -= force_j


@numba.njit
def outer_loop(positions, cell, offsets_list, neighbors_list, forces, energy, numbers):
    for ai in range(len(positions)):
        neighbors = neighbors_list[ai]
        offsets = offsets_list[ai]
        inner_loop(ai, positions, cell, offsets, neighbors, forces, energy, numbers)


class ZBLCalculator(Calculator):
    implemented_properties = ['energy', 'forces']

    def __init__(self, cutoff, skin=1.0, **kwargs):
        self._cutoff = cutoff
        self._skin = skin
        super().__init__(**kwargs)

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        if not ('forces' in properties or 'energy' in properties):
            return

        n_atoms = len(self.atoms)

        if 'numbers' in system_changes:
            self.nl = NeighborList([self._cutoff / 2] * n_atoms,
                                   self_interaction=False,
                                   skin=self._skin)

        self.nl.update(self.atoms)

        positions = self.atoms.positions
        numbers = self.atoms.numbers
        cell = self.atoms.cell.array

        neighbors_list, offsets_list = List(), List()
        for ai in range(n_atoms):
            neighbors, offsets = self.nl.get_neighbors(ai)
            neighbors_list.append(neighbors)
            offsets_list.append(offsets.astype(np.float64))

        energy = np.array([0.0])
        forces = np.zeros((n_atoms, 3))

        outer_loop(positions, cell, offsets_list, neighbors_list, forces, energy, numbers)

        self.results['energy'] = energy[0]
        self.results['forces'] = forces



from typing import Any, Dict, List
import numpy as np
from .input_output.pretty_table_prints import print_table
from pandas import DataFrame


class ClusterSpaceData:

    """Container class that holds information concerning a cluster space, including,
    e.g., the number of cluster, orbits, and parameters by order and number of bodies.

    Parameters
    ----------
    cs : ClusterSpace
        cluster space
    """

    def __init__(self, cs) -> None:
        self.max_nbody = cs.cutoffs.max_nbody
        self.max_order = cs.cutoffs.max_order

        # collect cutoff matrix
        self._cutoff_matrix = cs.cutoffs.cutoff_matrix

        # collect cluster, orbit, eigentensor counts
        self._cluster_counts = np.zeros((self.max_nbody, self.max_order - 1), dtype=int)
        self._orbit_counts = np.zeros((self.max_nbody, self.max_order - 1), dtype=int)
        self._eigentensor_counts = np.zeros((self.max_nbody, self.max_order - 1), dtype=int)
        for orbit in cs.orbits:
            proto_cluster = cs.cluster_list[orbit.prototype_index]
            order = len(proto_cluster)
            nbody = len(set(proto_cluster))
            self._cluster_counts[nbody-1, order-2] += len(orbit.orientation_families)
            self._orbit_counts[nbody-1, order-2] += 1
            self._eigentensor_counts[nbody-1, order-2] += len(orbit.eigentensors)

        # collect number of parameters after sum rule constraint
        self.ndofs_by_order = {o: cs.get_n_dofs_by_order(o) for o in cs.cutoffs.orders}

    def print_tables(self) -> None:
        """ Prints information concerning the underlying cluster space to stdout, including,
        e.g., the number of cluster, orbits, and parameters by order and number of bodies. """

        # print table data
        print('Cutoff Matrix')
        print_table(self._cutoff_matrix_padded)
        print('\nCluster counts')
        print_table(self._cluster_counts, include_sum=True)
        print('\nOrbit counts')
        print_table(self._orbit_counts, include_sum=True)
        print('\nEigentensor counts')
        print_table(self._eigentensor_counts, include_sum=True)

    @property
    def _cutoff_matrix_padded(self) -> np.array:
        """ Padded cutoff matrix with None for nbody=1 terms """
        return np.vstack(([[None] * (self.max_order - 1)], self._cutoff_matrix))

    def to_list(self) -> List[Dict[str, Any]]:
        """ Returns cluster space data in the form of a list of dicts. """
        records = []
        for order in range(2, self.max_order+1):
            for nbody in range(1, self.max_nbody+1):
                if nbody > order:
                    continue
                row = dict(order=order, nbody=nbody)
                row['cutoff'] = self._cutoff_matrix_padded[nbody-1, order-2]
                row['cluster_counts'] = self._cluster_counts[nbody-1, order-2]
                row['orbit_counts'] = self._orbit_counts[nbody-1, order-2]
                row['eigentensor_counts'] = self._eigentensor_counts[nbody-1, order-2]
                records.append(row)
        return records

    def to_dataframe(self) -> DataFrame:
        """ Returns cluster space data in the form of a pandas DataFrame. """
        return DataFrame.from_dict(self.to_list())

    def __str__(self) -> str:
        s = f'Cutoff matrix: \n {self._cutoff_matrix}\n'
        s += f'Cluster counts: \n{self._cluster_counts}\n'
        s += f'Orbit counts: \n{self._orbit_counts}\n'
        s += f'Eigentensor counts: \n{self._eigentensor_counts}\n'
        s += f'Degrees of freedoms: \n{self.ndofs_by_order}\n'
        return s

    def __repr__(self) -> str:
        return str(self)



"""
Contains the Orbit class which hold onformation about equivalent clusters.
"""
import numpy as np
import tarfile
import pickle

from .utilities import BiMap
from .atoms import Atom
from ..input_output.logging_tools import logger
from ..input_output.read_write_files import add_items_to_tarfile_pickle, read_items_pickle


logger = logger.getChild('orbits')


# This is the interface accessible for cluster_space
def get_orbits(cluster_list, atom_list, rotation_matrices, translation_vectors,
               permutations, prim, symprec):
    """Generate a list of the orbits for the clusters in a supercell
    configuration.

    This method requires as input a list of the clusters in a supercell
    configuration as well as a set of symmetry operations (rotations and
    translations). From this information it will generate a list of the orbits,
    i.e. the set of symmetry inequivalent clusters each associated with its
    respective set of equivalent clusters.

    Parameters
    ----------
    cluster_list : BiMap object
        a list of clusters
    atom_list : BiMap object
        a list of atoms in a supercell
    rotation_matrices : list of NumPy (3,3) arrays
        rotational symmetries to be imposed (e.g., from spglib)
    translation_vectors : list of NumPy (3) arrays
        translational symmetries to be imposed (e.g., from spglib)
    permutations : list of permutations
        lookup table for permutations
    prim : hiPhive Atoms object
        primitive structure

    Returns
    -------
    list of Orbits objs
        orbits associated with the list of input clusters

    """

    logger.debug('Preparing input...')
    atoms = prepare_atoms(atom_list)
    clusters = prepare_clusters(cluster_list)
    rotations = prepare_rotations(rotation_matrices)
    translations = prepare_translations(translation_vectors)
    permutations = prepare_permutations(permutations)
    cell = prim.cell
    basis = prim.basis

    logger.debug('Creating permutation map...')
    permutation_map, extended_atoms = \
        get_permutation_map(atoms, rotations, translations, basis, symprec)

    logger.debug('Creating orbits...')
    orbits = _get_orbits(permutation_map, extended_atoms, clusters, basis,
                         cell, rotations, permutations)

    return orbits


# All prepares are in case we changes some interface stuff
def prepare_permutations(permutations):
    perms = BiMap()
    for p in permutations:
        perms.append(tuple(p))
    return perms


def prepare_atoms(atom_list):
    atoms = BiMap()
    for atom in atom_list:
        atoms.append(atom)
    return atoms


def prepare_clusters(cluster_list):
    clusters = BiMap()
    for cluster in cluster_list:
        clusters.append(tuple(cluster))
    return clusters


def prepare_rotations(rotation_matrices):
    return rotation_matrices


def prepare_translations(translation_vectors):
    return translation_vectors


def get_permutation_map(atoms, rotations, translations, basis, symprec):

    extended_atoms = atoms.copy()

    permutation_map = np.zeros((len(atoms), len(rotations)), dtype=int)

    scaled_positions = [atom.spos(basis) for atom in extended_atoms]

    for sym_index, (R, T) in enumerate(zip(rotations, translations)):
        for atom_index, spos in enumerate(scaled_positions):

            new_spos = np.dot(R, spos) + T
            new_atom = Atom.spos_to_atom(new_spos, basis, symprec)

            if new_atom not in extended_atoms:
                extended_atoms.append(new_atom)

            mapped_atom_index = extended_atoms.index(new_atom)
            permutation_map[atom_index, sym_index] = mapped_atom_index

    return permutation_map, extended_atoms


# The internal function doing the outer loop over orbits
def _get_orbits(permutation_map, extended_atoms, clusters,
                basis, cell,
                rotations, permutations):
    cluster_is_found = [False] * len(clusters)
    orbits = []
    for cluster_index, cluster in enumerate(clusters):
        if cluster_is_found[cluster_index]:
            continue

        orbit = Orbit()

        cluster_atoms = [extended_atoms[i] for i in cluster]

        positions = [atom.pos(basis, cell) for atom in cluster_atoms]
        orbit.radius = get_geometrical_radius(positions)
        orbit.maximum_distance = get_maximum_distance(positions)
        orbit.order = len(cluster)

        populate_orbit(orbit, permutations, clusters,
                       cluster,
                       permutation_map, extended_atoms,
                       cluster_is_found)
        orbits.append(orbit)

#            bar.tick()
    return orbits


# Takes a cluster and generates all equivalent, translated clusters
def generate_translated_clusters(cluster, extended_atoms):
    transformed_cluster_atoms = [extended_atoms[i] for i in cluster]
    tested_offsets = set()
    for atom in transformed_cluster_atoms:
        offset = atom.offset
        if offset in tested_offsets:
            continue
        else:
            tested_offsets.add(offset)
        translated_cluster = []
        for atom in transformed_cluster_atoms:
            new_offset = np.subtract(atom.offset, offset)
            new_atom = Atom(atom.site, new_offset)
            translated_cluster.append(extended_atoms.index(new_atom))
        yield tuple(translated_cluster)


# Here is the actual categorization
def populate_orbit(orbit, permutations, clusters,
                   cluster,
                   permutation_map, extended_atoms,
                   cluster_is_found):
    for sym_index in range(permutation_map.shape[1]):

        of = OrientationFamily(sym_index)

        transformed_cluster = permutation_map[cluster, sym_index]

        for translated_cluster in generate_translated_clusters(
                transformed_cluster, extended_atoms):

            argsort = tuple(np.argsort(translated_cluster))
            perm_index = permutations.index(argsort)

            translated_cluster = tuple(sorted(translated_cluster))
            translated_cluster_index = clusters.index(translated_cluster)

            if cluster == translated_cluster:
                if (sym_index, perm_index) not in orbit.eigensymmetries:
                    orbit.eigensymmetries.append((sym_index, perm_index))

            if not cluster_is_found[translated_cluster_index]:
                cluster_is_found[translated_cluster_index] = True
                of.cluster_indices.append(translated_cluster_index)
                of.permutation_indices.append(perm_index)

        if len(of.cluster_indices) > 0:
            orbit.orientation_families.append(of)

    return orbit


class Orbit:
    """
    This class serves as a container for storing data pertaining to an orbit.

    Attributes
    ----------
    orientation_families : list of OrientationFamily objs
        orientation families of the orbit
    eigensymmetries : list of tuples
        each eigensymmetry corresponds to a pair where the first index
        is the symmetry and the second is the permutation
    eigentensors : list(numpy.ndarray)
        decomposition of the force constant into symmetry elements
    """

    # TODO: Make properties of the parameters
    def __init__(self):
        self.orientation_families = []
        self.eigensymmetries = []
        self.eigentensors = []
        self.radius = None
        self.order = None
        self.maximum_distance = None

    @property
    def prototype_index(self):
        """int : index of cluster that serves as prototype for this orbit

        In the code the first symmetry is always the identity so the first
        orientation family should always correspond to the prototype"""
        return self.orientation_families[0].cluster_indices[0]

    def write(self, f):
        """Write a Orbit instance to a file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        tar_file = tarfile.open(mode='w', fileobj=f)

        # add all available attributes
        add_items_to_tarfile_pickle(tar_file, self.__dict__, 'attributes')

    @staticmethod
    def read(f):
        """Load a ClusterSpace from file

        Parameters
        ----------
        f : string or file object
            name of input file (string) or stream to load from (file object)
        """

        orb = Orbit()
        tar_file = tarfile.open(mode='r', fileobj=f)

        # read attributes pickle
        attributes = read_items_pickle(tar_file, 'attributes')
        for name, value in attributes.items():
            orb.__setattr__(name, value)
        return orb


class OrientationFamily:
    """A container for storing information for a "family of orientations".

    An orbit contains many clusters. Some of the clusters can be tranlsated
    onto each other and other must first be rotated. A set of clusters in the
    orbit which can all be translated onto each other are oriented in the same
    way and belongs to the same orientation family. The family is haracterized
    by the symmetry (rotation) which relates it to the prototype structure of
    the orbit.

    Since the clusters are generally stored sorted the permutation information
    must also be stored.

    Parameters
    ----------
    symmetry_index : int
        The index of the symmetry corresponding to spglibs symmetry

    Attributes
    ----------
    symmetry_index : int
        The index of the symmetry corresponding to spglibs symmetry
    cluster_indices : list of ints
        The indices of the clusters belonging to this family
    permutation_indices : list of ints
        The indices of the permutation vector
    """

    def __init__(self, symmetry_index=None):
        self.symmetry_index = symmetry_index
        self.cluster_indices = []
        self.permutation_indices = []

    def write(self, f):
        """ Write the object to file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        pickle.dump(self, f)

    @staticmethod
    def read(f):
        """Load a OrientationFamily object from a pickle file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        OrientationFamily object
        """
        return pickle.load(f)


def get_geometrical_radius(positions):
    """Compute the geometrical size of a 3-dimensional point cloud. The
    geometrical size is defined as the average distance to the geometric
    center.

    Parameters
    ----------
    positions : list of 3-dimensional vectors
        positions of points in cloud

    Returns
    -------
    float
        geometric size of point cloud
    """
    geometric_center = np.mean(positions, axis=0)
    return np.mean(np.sqrt(np.sum((positions - geometric_center)**2, axis=1)))


def get_maximum_distance(positions):
    """Compute the maximum distance between any two points in a 3-dimensional
    point cloud. This is equivalent to the "size" criterion used when imposing
    a certain (pair) cutoff criterion during construction of a set of clusters.

    Parameters
    ----------
    positions : list of 3-dimensional vectors
        positions of points in cloud

    Returns
    -------
    float
        maximum distance betwee any two points
    """
    if len(positions) == 0:
        return 0.0
    max_distance = 0.0
    for pt1 in positions[:-1]:
        for pt2 in positions[1:]:
            max_distance = max(max_distance, np.linalg.norm(pt1 - pt2))
    return max_distance



"""
Functionality for enforcing rotational sum rules
"""
import numpy as np
from typing import List
from sklearn.linear_model import Ridge
from scipy.sparse import coo_matrix, lil_matrix
from ..cluster_space import ClusterSpace


def enforce_rotational_sum_rules(cs: ClusterSpace,
                                 parameters: np.ndarray,
                                 sum_rules: List[str] = None,
                                 alpha: float = 1e-6,
                                 **ridge_kwargs: dict) -> np.ndarray:
    """ Enforces rotational sum rules by projecting parameters.

    Note
    ----
    The interface to this function might change in future releases.

    Parameters
    ----------
    cs
        the underlying cluster space
    parameters
        parameters to be constrained
    sum_rules
        type of sum rules to enforce; possible values: 'Huang', 'Born-Huang'
    alpha
        hyperparameter to the ridge regression algorithm; keyword argument
        passed to the optimizer; larger values specify stronger regularization,
        i.e., less correction but higher stability [default: 1e-6]
    ridge_kwargs
        kwargs to be passed to sklearn Ridge

    Returns
    -------
    numpy.ndarray
        constrained parameters

    Examples
    --------
    The rotational sum rules can be enforced to the parameters before
    constructing a force constant potential as illustrated by the following
    snippet::

        cs = ClusterSpace(reference_structure, cutoffs)
        sc = StructureContainer(cs)
        # add structures to structure container
        opt = Optimizer(sc.get_fit_data())
        opt.train()
        new_params = enforce_rotational_sum_rules(cs, opt.parameters,
            sum_rules=['Huang', 'Born-Huang'])
        fcp = ForceConstantPotential(cs, new_params)

    """

    all_sum_rules = ['Huang', 'Born-Huang']

    # setup
    parameters = parameters.copy()
    if sum_rules is None:
        sum_rules = all_sum_rules

    # get constraint-matrix
    M = get_rotational_constraint_matrix(cs, sum_rules)

    # before fit
    d = M.dot(parameters)
    delta = np.linalg.norm(d)
    print('Rotational sum-rules before, ||Ax|| = {:20.15f}'.format(delta))

    # fitting
    ridge = Ridge(alpha=alpha, fit_intercept=False, solver='sparse_cg', **ridge_kwargs)
    ridge.fit(M, d)
    parameters -= ridge.coef_

    # after fit
    d = M.dot(parameters)
    delta = np.linalg.norm(d)
    print('Rotational sum-rules after,  ||Ax|| = {:20.15f}'.format(delta))

    return parameters


def get_rotational_constraint_matrix(cs, sum_rules=None):

    all_sum_rules = ['Huang', 'Born-Huang']

    if sum_rules is None:
        sum_rules = all_sum_rules

    # setup
    assert len(sum_rules) > 0
    for s in sum_rules:
        if s not in all_sum_rules:
            raise ValueError('Sum rule {} not allowed, select from {}'.format(s, all_sum_rules))

    # make orbit-parameter index map
    params = _get_orbit_parameter_map(cs)
    lookup = _get_fc_lookup(cs)

    # append the sum rule matrices
    Ms = []
    args = (lookup, params, cs.atom_list, cs._prim)
    for sum_rule in sum_rules:
        if sum_rule == 'Huang':
            Ms.append(_create_Huang_constraint(*args))
        elif sum_rule == 'Born-Huang':
            Ms.append(_create_Born_Huang_constraint(*args))

    # transform and stack matrices
    cvs_trans = cs._cvs
    for i, M in enumerate(Ms):
        M = M.dot(cvs_trans)
        M = M.toarray()
        Ms[i] = M

    return np.vstack(Ms)


def _get_orbit_parameter_map(cs):
    # make orbit-parameter index map
    params = []
    n = 0
    for orbit_index, orbit in enumerate(cs.orbits):
        n_params_in_orbit = len(orbit.eigentensors)
        params.append(list(range(n, n + n_params_in_orbit)))
        n += n_params_in_orbit
    return params


def _get_fc_lookup(cs):
    # create lookuptable for force constants
    lookup = {}
    for orbit_index, orbit in enumerate(cs.orbits):
        for of in orbit.orientation_families:
            for cluster_index, perm_index in zip(of.cluster_indices, of.permutation_indices):
                cluster = cs._cluster_list[cluster_index]
                perm = cs._permutations[perm_index]
                lookup[tuple(cluster)] = [et.transpose(perm) for et in of.eigentensors], orbit_index
    return lookup


def _create_Huang_constraint(lookup, parameter_map, atom_list, prim):

    m = np.zeros((parameter_map[-1][-1] + 1, 3**4))

    def R(i, j):
        pi = atom_list[i].pos(prim.basis, prim.cell)
        pj = atom_list[j].pos(prim.basis, prim.cell)
        return pi - pj

    for i in range(len(prim)):
        for j in range(len(atom_list)):
            ets, orbit_index = lookup.get(tuple(sorted((i, j))), (None, None))
            if ets is None:
                continue
            inv_perm = np.argsort(np.argsort((i, j)))
            et_indices = parameter_map[orbit_index]
            for et, et_index in zip(ets, et_indices):
                et = et.transpose(inv_perm)
                Rij = R(i, j)
                Cij = np.einsum(et, [0, 1], Rij, [2], Rij, [3])
                Cij -= Cij.transpose([2, 3, 0, 1])
                m[et_index] += Cij.flat

    m = coo_matrix(m.transpose())
    return m


def _create_Born_Huang_constraint(lookup, parameter_map, atom_list, prim):

    # Use scipy list-of-lists sparse matrix for good tradeoff between
    # restructuring, indexing/slicing and memory footprint
    M = lil_matrix((len(prim) * 3**3, parameter_map[-1][-1] + 1))
    for i in range(len(prim)):
        # Use smaller numpy arrays for speedy arithmetic
        m = np.zeros((parameter_map[-1][-1] + 1, 3**3))
        for j in range(len(atom_list)):
            ets, orbit_index = lookup.get(tuple(sorted((i, j))), (None, None))
            if ets is None:
                continue
            inv_perm = np.argsort(np.argsort((i, j)))
            et_indices = parameter_map[orbit_index]
            R = atom_list[j].pos(prim.basis, prim.cell)
            for et, et_index in zip(ets, et_indices):
                et = et.transpose(inv_perm)
                tmp = np.einsum(et, [0, 1], R, [2])
                tmp -= tmp.transpose([0, 2, 1])
                m[et_index] += tmp.flat
        M[i*3**3:(i+1)*3**3, :] = m.transpose()

    # Convert lil_matrix to coo_matrix
    return M.tocoo()



"""
This module contains various support/utility functions.
"""

from typing import List, Tuple
import numpy as np

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from ase.geometry import find_mic
from ase.geometry import get_distances
from ase.neighborlist import neighbor_list
from .cluster_space import ClusterSpace
from .force_constants import ForceConstants
from .input_output.logging_tools import logger


logger = logger.getChild('utilities')


def get_displacements(atoms: Atoms,
                      atoms_ideal: Atoms,
                      cell_tol: float = 1e-4) -> np.ndarray:
    """Returns the the smallest possible displacements between a
    displaced configuration relative to an ideal (reference)
    configuration.

    Notes
    -----
    * uses :func:`ase.geometry.find_mic`
    * assumes periodic boundary conditions in all directions

    Parameters
    ----------
    atoms
        configuration with displaced atoms
    atoms_ideal
        ideal configuration relative to which displacements are computed
    cell_tol
        cell tolerance; if cell missmatch more than tol value error is raised
    """
    if not np.array_equal(atoms.numbers, atoms_ideal.numbers):
        raise ValueError('Atomic numbers do not match.')
    if np.linalg.norm(atoms.cell - atoms_ideal.cell) > cell_tol:
        raise ValueError('Cells do not match.')

    raw_position_diff = atoms.positions - atoms_ideal.positions
    wrapped_mic_displacements = find_mic(raw_position_diff, atoms_ideal.cell, pbc=True)[0]
    return wrapped_mic_displacements


def _get_forces_from_atoms(atoms: Atoms, calc=None) -> np.ndarray:
    """ Try to get forces from an atoms object """

    # Check if two calculators are available
    if atoms.calc is not None and calc is not None:
        raise ValueError('Atoms.calc is not None and calculator was provided')

    # If calculator is provided as argument
    if calc is not None:
        atoms_tmp = atoms.copy()
        atoms_tmp.calc = calc
        forces_calc = atoms_tmp.get_forces()
        if 'forces' in atoms.arrays:
            if not np.allclose(forces_calc, atoms.get_array('forces')):
                raise ValueError('Forces in atoms.arrays are different from the calculator forces')
        return forces_calc

    # If calculator is attached
    if atoms.calc is not None:
        if not isinstance(atoms.calc, SinglePointCalculator):
            raise ValueError('atoms.calc is not a SinglePointCalculator')

        forces_calc = atoms.get_forces()
        if 'forces' in atoms.arrays:
            if not np.allclose(forces_calc, atoms.get_array('forces')):
                raise ValueError('Forces in atoms.arrays are different from the calculator forces')
        return forces_calc

    # No calculator attached or provided as argument, forces should therefore be in atoms.arrays
    if 'forces' in atoms.arrays:
        forces = atoms.get_array('forces')
    else:
        raise ValueError('Unable to find forces')
    return forces


def prepare_structure(atoms: Atoms,
                      atoms_ideal: Atoms,
                      calc: SinglePointCalculator = None,
                      check_permutation: bool = True) -> Atoms:
    """Prepare a structure in the format suitable for a
    :class:`StructureContainer <hiphive.StructureContainer>`.

    Either forces should be attached to input atoms object as an array,
    or the atoms object should have a SinglePointCalculator attached to it containing
    forces, or a calculator (calc) should be supplied.

    Parameters
    ----------
    atoms
        input structure
    atoms_ideal
        reference structure relative to which displacements are computed
    check_permutation
        whether find_permutation should be used or not
    calc
        ASE calculator used for computing forces

    Returns
    -------
    ASE atoms object
        prepared ASE atoms object with forces and displacements as arrays
    """
    # get forces
    forces = _get_forces_from_atoms(atoms, calc=calc)

    # setup new atoms
    if check_permutation:
        perm = find_permutation(atoms, atoms_ideal)
    else:
        perm = np.array([f for f in range(len(atoms))])
    atoms_new = atoms.copy()
    atoms_new = atoms_new[perm]
    atoms_new.arrays['forces'] = forces[perm]
    disps = get_displacements(atoms_new, atoms_ideal)
    atoms_new.arrays['displacements'] = disps
    atoms_new.positions = atoms_ideal.positions

    return atoms_new


def prepare_structures(structures: List[Atoms],
                       atoms_ideal: Atoms,
                       calc: SinglePointCalculator = None,
                       check_permutation: bool = True) -> List[Atoms]:
    """Prepares a set of structures in the format suitable for adding them to
    a :class:`StructureContainer <hiphive.StructureContainer>`.

    `structures` should represent a list of supercells with displacements
    while `atoms_ideal` should provide the ideal reference structure (without
    displacements) for the given structures.

    The structures that are returned will have their positions reset to the
    ideal structures. Displacements and forces will be added as arrays to the
    atoms objects.

    If no calculator is provided, then there must be an ASE
    `SinglePointCalculator <ase.calculators.singlepoint>` object attached to
    the structures or the forces should already be attached as
    arrays to the structures.

    If a calculator is provided then it will be used to compute the forces for
    all structures.

    Example
    -------

    The following example illustrates the use of this function::

        db = connect('dft_training_structures.db')
        training_structures = [row.toatoms() for row in db.select()]
        training_structures = prepare_structures(training_structures, atoms_ideal)
        for s in training_structures:
            sc.add_structure(s)

    Parameters
    ----------
    structures
        list of input displaced structures
    atoms_ideal
        reference structure relative to which displacements are computed
    calc
        ASE calculator used for computing forces

    Returns
    -------
    list of prepared structures with forces and displacements as arrays
    """
    return [prepare_structure(s, atoms_ideal, calc, check_permutation) for s in structures]


def find_permutation(atoms: Atoms, atoms_ref: Atoms) -> List[int]:
    """ Returns the best permutation of atoms for mapping one
    configuration onto another.

    Parameters
    ----------
    atoms
        configuration to be permuted
    atoms_ref
        configuration onto which to map

    Examples
    --------
    After obtaining the permutation via ``p = find_permutation(atoms1, atoms2)``
    the reordered structure ``atoms1[p]`` will give the closest match
    to ``atoms2``.
    """
    assert np.linalg.norm(atoms.cell - atoms_ref.cell) < 1e-6
    permutation = []
    for i in range(len(atoms_ref)):
        dist_row = get_distances(
            atoms.positions, atoms_ref.positions[i], cell=atoms_ref.cell, pbc=True)[1][:, 0]
        permutation.append(np.argmin(dist_row))

    if len(set(permutation)) != len(permutation):
        raise Exception('Duplicates in permutation')
    for i, p in enumerate(permutation):
        if atoms[p].symbol != atoms_ref[i].symbol:
            raise Exception('Matching lattice sites have different occupation')
    return permutation


class Shell:
    """
    Neighbor Shell class

    Parameters
    ----------
    types : list or tuple
        atomic types for neighbor shell
    distance : float
        interatomic distance for neighbor shell
    count : int
        number of pairs in the neighbor shell
    """

    def __init__(self,
                 types: List[str],
                 distance: float,
                 count: int = 0):
        self.types = types
        self.distance = distance
        self.count = count

    def __str__(self):
        s = '{}-{}   distance: {:10.6f}    count: {}'.format(*self.types, self.distance, self.count)
        return s

    __repr__ = __str__


def get_neighbor_shells(atoms: Atoms,
                        cutoff: float,
                        dist_tol: float = 1e-5) -> List[Shell]:
    """ Returns a list of neighbor shells.

    Distances are grouped into shells via the following algorithm:

    1. Find smallest atomic distance `d_min`

    2. Find all pair distances in the range `d_min + 1 * dist_tol`

    3. Construct a shell from these and pop them from distance list

    4. Go to 1.

    Parameters
    ----------
    atoms
        configuration used for finding shells
    cutoff
        exclude neighbor shells which have a distance larger than this value
    dist_tol
        distance tolerance
    """

    # get distances
    ijd = neighbor_list('ijd', atoms, cutoff)
    ijd = list(zip(*ijd))
    ijd.sort(key=lambda x: x[2])

    # sort into shells
    symbols = atoms.get_chemical_symbols()
    shells = []
    for i, j, d in ijd:
        types = tuple(sorted([symbols[i], symbols[j]]))
        for shell in shells:
            if abs(d - shell.distance) < dist_tol and types == shell.types:
                shell.count += 1
                break
        else:
            shell = Shell(types, d, 1)
            shells.append(shell)
    shells.sort(key=lambda x: (x.distance, x.types, x.count))

    # warning if two shells are within 2 * tol
    for i, s1 in enumerate(shells):
        for j, s2 in enumerate(shells[i+1:]):
            if s1.types != s2.types:
                continue
            if not s1.distance < s2.distance - 2 * dist_tol:
                logger.warning('Found two shells within 2 * dist_tol')

    return shells


def extract_parameters(fcs: ForceConstants,
                       cs: ClusterSpace,
                       sanity_check: bool = True,
                       lstsq_method: str = 'numpy') \
        -> Tuple[np.ndarray, np.ndarray, int, np.ndarray]:
    """ Extracts parameters from force constants.


    This function can be used to extract parameters to create a
    ForceConstantPotential from a known set of force constants.
    The return values come from NumPy's `lstsq function
    <https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.lstsq.html>`_
    or from SciPy's `sparse lsqr function
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsqr.html>`_.
    Using `lstsq_method='scipy'` might be faster and have a smaller memory footprint for large
    systems, at the expense of some accuracy. This is due to the use of sparse matrices
    and an iterative solver.

    Parameters
    ----------
    fcs
        force constants
    cs
        cluster space
    sanity_check
        bool whether or not to perform a sanity check by computing the relative error between
        the input fcs and the output fcs
    lstsq_method
        method to use when making a least squares fit of a ForceConstantModel to the given fcs,
        allowed values are 'numpy' for `np.linalg.lstsq` or 'scipy' `for scipy.sparse.linalg.lsqr`

    Returns
    -------
    parameters
        parameters that together with the ClusterSpace generates the best representation of the FCs

    """
    from .force_constant_model import ForceConstantModel
    from .force_constant_potential import ForceConstantPotential
    from scipy.sparse.linalg import lsqr

    if lstsq_method not in ['numpy', 'scipy']:
        raise ValueError('lstsq_method must be either numpy or scipy')

    # extract the parameters
    fcm = ForceConstantModel(fcs.supercell, cs)
    # If the cluster space large, a sparse least squares solver is faster
    if lstsq_method == 'numpy':
        A, b = fcm.get_fcs_sensing(fcs, sparse=False)
        parameters = np.linalg.lstsq(A, b, rcond=None)[0]
    elif lstsq_method == 'scipy':
        A, b = fcm.get_fcs_sensing(fcs, sparse=True)
        # set minimal tolerances to maximize iterative least squares accuracy
        parameters = lsqr(A, b, atol=0, btol=0, conlim=0)[0]

    # calculate the relative force constant error
    if sanity_check:
        fcp = ForceConstantPotential(cs, parameters)
        fcs_hiphive = fcp.get_force_constants(fcs.supercell)
        for order in cs.cutoffs.orders:
            fc_original = fcs.get_fc_array(order=order)
            fc_reconstructed = fcs_hiphive.get_fc_array(order=order)
            rel_error = np.linalg.norm(fc_original-fc_reconstructed) / np.linalg.norm(fc_original)
            print(f'Force constant reconstruction error order {order}: {100*rel_error:9.4f}%')

    return parameters



"""
hiPhive module.
"""

from .cluster_space import ClusterSpace
from .structure_container import StructureContainer
from .force_constant_potential import ForceConstantPotential
from .force_constants import ForceConstants
from .core.config import config
from .core.rotational_constraints import enforce_rotational_sum_rules

__project__ = 'hiPhive'
__description__ = 'High-order force constants for the masses'
__authors__ = ['Fredrik Eriksson',
               'Erik Fransson',
               'Paul Erhart']
__copyright__ = '2024'
__license__ = 'MIT License'
__credits__ = ['Fredrik Eriksson',
               'Erik Fransson',
               'Paul Erhart']
__version__ = '1.4'
__all__ = ['ClusterSpace',
           'StructureContainer',
           'ForceConstantPotential',
           'ForceConstants',
           'config',
           'enforce_rotational_sum_rules']
__maintainer__ = 'The hiPhive developers team'
__maintainer_email__ = 'hiphive@materialsmodeling.org'
__status__ = 'Stable'
__url__ = 'http://hiphive.materialsmodeling.org/'



"""
The ``io.shengBTE`` module provides functions for reading and writing data
files in shengBTE format.
"""

import numpy as np
from itertools import product
from collections import namedtuple
from ..core.structures import Supercell
import hiphive


def read_shengBTE_fc3(filename, prim, supercell, symprec=1e-5):
    """Reads third-order force constants file in shengBTE format.

    Parameters
    ----------
    filename : str
        input file name
    prim : ase.Atoms
        primitive cell for the force constants
    supercell : ase.Atoms
        supercell onto which to map force constants
    symprec : float
        structural symmetry tolerance

    Returns
    -------
    fcs : ForceConstants
        third order force constants for the specified supercell
    """

    raw_sheng = _read_raw_sheng(filename)

    sheng = _raw_to_fancy(raw_sheng, prim.cell)

    fcs = _sheng_to_fcs(sheng, prim, supercell, symprec)

    return fcs


def write_shengBTE_fc3(filename, fcs, prim, symprec=1e-5, cutoff=np.inf,
                       fc_tol=1e-8):
    """Writes third-order force constants file in shengBTE format.

    Parameters
    -----------
    filename : str
        input file name
    fcs : ForceConstants
        force constants; the supercell associated with this object must be
        based on prim
    prim : ase.Atoms
        primitive configuration (must be equivalent to structure used in the
        shengBTE calculation)
    symprec : float
        structural symmetry tolerance
    cutoff : float
        all atoms in cluster must be within this cutoff
    fc_tol : float
        if the absolute value of the largest entry in a force constant is less
        than fc_tol it will not be written
    """

    sheng = _fcs_to_sheng(fcs, prim, symprec, cutoff, fc_tol)

    raw_sheng = _fancy_to_raw(sheng)

    _write_raw_sheng(raw_sheng, filename)


def _read_raw_sheng(filename):
    """ Read shengBTE fc3 file.

    Parameters
    ----------
    filename : str
        input file name

    Returns
    -------
    list
        list with shengBTE block entries, where an entry consist of
        [i, j, k, cell_pos2, cell_pos3, fc3]
    """
    lines_per_fc_block = 32

    fc3_data_shengBTE = []
    with open(filename, 'r') as f:
        lines = f.readlines()
        num_fcs = int(lines[0])
        lind = 2
        for i in range(1, num_fcs+1):
            # sanity check
            assert int(lines[lind]) == i, (int(lines[lind]), i)

            # read cell poses
            cell_pos2 = np.array([float(fld) for fld in lines[lind+1].split()])
            cell_pos3 = np.array([float(fld) for fld in lines[lind+2].split()])

            # read basis indices
            i, j, k = (int(fld) for fld in lines[lind+3].split())

            # read fc
            fc3_ijk = np.zeros((3, 3, 3))
            for n in range(27):
                x, y, z = [int(fld) - 1 for fld in lines[lind+4+n].split()[:3]]
                fc3_ijk[x, y, z] = float(lines[lind+4+n].split()[-1])

            # append entry
            entry = [i, j, k, cell_pos2, cell_pos3, fc3_ijk]
            fc3_data_shengBTE.append(entry)
            lind += lines_per_fc_block

    return fc3_data_shengBTE


_ShengEntry = namedtuple('Entry', ['site_0', 'site_1', 'site_2', 'pos_1',
                                   'pos_2', 'fc', 'offset_1', 'offset_2'])


def _raw_to_fancy(raw_sheng, cell):
    """
    Converts force constants as read from shengBTE file to the namedtuple
    format defined above (_ShengEntry).
    """
    sheng = []
    for raw_entry in raw_sheng:
        p1, p2 = raw_entry[3:5]
        offset_1 = np.linalg.solve(cell.T, p1).round(0).astype(int)
        offset_2 = np.linalg.solve(cell.T, p2).round(0).astype(int)
        entry = _ShengEntry(*(i - 1 for i in raw_entry[:3]), *raw_entry[3:],
                            offset_1, offset_2)
        sheng.append(entry)
    return sheng


def _fancy_to_raw(sheng):
    """
    Converts force constants namedtuple format defined above (_ShengEntry) to
    format used for writing shengBTE files.
    """
    raw_sheng = []
    for entry in sheng:
        raw_entry = list(entry[:6])
        raw_entry[0] += 1
        raw_entry[1] += 1
        raw_entry[2] += 1
        raw_sheng.append(raw_entry)

    return raw_sheng


def _write_raw_sheng(raw_sheng, filename):
    """ See corresponding read function. """

    with open(filename, 'w') as f:
        f.write('{}\n\n'.format(len(raw_sheng)))

        for index, fc3_row in enumerate(raw_sheng, start=1):
            i, j, k, cell_pos2, cell_pos3, fc3_ijk = fc3_row

            f.write('{:5d}\n'.format(index))

            f.write((3*'{:14.10f} '+'\n').format(*cell_pos2))
            f.write((3*'{:14.10f} '+'\n').format(*cell_pos3))
            f.write((3*'{:5d}'+'\n').format(i, j, k))

            for x, y, z in product(range(3), repeat=3):
                f.write((3*' {:}').format(x+1, y+1, z+1))
                f.write('    {:14.10f}\n'.format(fc3_ijk[x, y, z]))
            f.write('\n')


def _fcs_to_sheng(fcs, prim, symprec, cutoff, fc_tol):
    """ Produces a list containing fcs in shengBTE format

    prim and fcs.supercell must be aligned.
    """

    supercell = Supercell(fcs.supercell, prim, symprec)
    assert all(fcs.supercell.pbc) and all(prim.pbc)

    n_atoms = len(supercell)

    D = fcs.supercell.get_all_distances(mic=False, vector=True)
    D_mic = fcs.supercell.get_all_distances(mic=True, vector=True)
    M = np.eye(n_atoms, dtype=bool)
    for i in range(n_atoms):
        for j in range(i + 1, n_atoms):
            M[i, j] = (np.allclose(D[i, j], D_mic[i, j], atol=symprec, rtol=0)
                       and np.linalg.norm(D[i, j]) < cutoff)
            M[j, i] = M[i, j]

    data = {}
    for a0 in supercell:
        for a1 in supercell:
            if not M[a0.index, a1.index]:
                continue
            for a2 in supercell:
                if not (M[a0.index, a2.index] and M[a1.index, a2.index]):
                    continue

                offset_1 = np.subtract(a1.offset, a0.offset)
                offset_2 = np.subtract(a2.offset, a0.offset)

                sites = (a0.site, a1.site, a2.site)

                key = sites + tuple(offset_1) + tuple(offset_2)

                ijk = (a0.index, a1.index, a2.index)

                fc = fcs[ijk]

                if key in data:
                    assert np.allclose(data[key], fc, atol=fc_tol)
                else:
                    data[key] = fc

    sheng = []
    for k, fc in data.items():
        if np.max(np.abs(fc)) < fc_tol:
            continue
        offset_1 = k[3:6]
        pos_1 = np.dot(offset_1, prim.cell)
        offset_2 = k[6:9]
        pos_2 = np.dot(offset_2, prim.cell)
        entry = _ShengEntry(*k[:3], pos_1, pos_2, fc, offset_1, offset_2)
        sheng.append(entry)

    return sheng


def _sheng_to_fcs(sheng, prim, supercell, symprec):
    supercell_map = Supercell(supercell, prim, symprec)
    fc_array = np.zeros((len(supercell),) * 3 + (3,) * 3)
    mapped_clusters = np.zeros((len(supercell),) * 3, dtype=bool)

    for atom in supercell_map:
        i = atom.index
        for entry in sheng:
            if atom.site != entry.site_0:
                continue
            j = supercell_map.index(entry.site_1, entry.offset_1 + atom.offset)
            k = supercell_map.index(entry.site_2, entry.offset_2 + atom.offset)
            ijk = (i, j, k)
            if mapped_clusters[ijk]:
                raise Exception('Ambiguous force constant.'
                                ' Supercell is too small')
            fc_array[ijk] = entry.fc
            mapped_clusters[ijk] = True

    fcs = hiphive.force_constants.ForceConstants.from_arrays(
        supercell, fc3_array=fc_array)
    return fcs



"""
This module provides functionality for storing and handling of force constants.
"""

import numpy as np
import tarfile

from abc import ABC, abstractmethod
from itertools import product
from string import ascii_lowercase, ascii_uppercase
from typing import List, Tuple

from scipy.linalg import eig
from ase import Atoms, units

from .input_output import shengBTE as shengBTEIO
from .input_output import phonopy as phonopyIO
from .input_output.read_write_files import add_items_to_tarfile_pickle, read_items_pickle, \
    add_ase_atoms_to_tarfile, read_ase_atoms_from_tarfile
from .input_output import gpumd as gpumdIO


class ForceConstants(ABC):
    """ Base class for force constants """

    def __init__(self, supercell: Atoms):
        if not all(supercell.pbc):
            raise ValueError('configuration must have periodic boundary conditions')
        self._supercell = supercell.copy()

    @abstractmethod
    def __getitem__(self):
        pass

    @property
    def n_atoms(self) -> int:
        """ number of atoms """
        return len(self.supercell)

    @property
    def supercell(self) -> Atoms:
        """ supercell associated with force constants """
        return self._supercell.copy()

    @property
    def clusters(self) -> list:
        """ sorted list of clusters """
        return sorted(self._fc_dict.keys(), key=lambda c: (len(c), c))

    def __len__(self) -> int:
        """ int : number of clusters (or force constants) """
        return len(self._fc_dict)

    def __add__(self, fcs_other) -> None:
        """ ForceConstants : addition of two force constants """
        if not isinstance(fcs_other, type(self)):
            raise ValueError('ForceConstants objects are of different types')

        # if they have overlapping orders, raise
        if any(order in self.orders for order in fcs_other.orders):
            raise ValueError('ForceConstants objects share at least one order')

        # check that they share the same supercell
        if self.supercell != fcs_other.supercell:
            raise ValueError('supercells do not match')

        # add force constants
        fc_merged = {**self._fc_dict, **fcs_other._fc_dict}
        return self.__class__(fc_merged, self.supercell)

    def get_fc_dict(self, order: int = None) -> dict:
        """ Returns force constant dictionary for one specific order.

        The returned dict may be sparse or may be dense depending on the
        underlying force constants.

        Parameters
        ----------
        order
            force constants returned for this order

        Returns
        -------
        dictionary with keys corresponding to clusters and values to
        respective force constant
        """
        if order is None:
            return self._fc_dict

        if order not in self.orders:
            raise ValueError('Order {} not in ForceConstants'.format(order))

        fc_order = {}
        for c, fc in self._fc_dict.items():
            if len(c) == order:
                fc_order[c] = fc
        return fc_order

    def get_fc_array(self, order: int, format: str = 'phonopy') -> np.ndarray:
        """ Returns force constants in array format for specified order.

        Parameters
        ----------
        order
            force constants for this order will be returned
        format
            specify which format (shape) the NumPy array should have,
            possible values are `phonopy` and `ase`

        Returns
        -------
        NumPy array with shape `(N,)*order + (3,)*order` where `N` is
        the number of atoms
        """
        if order not in self.orders:
            raise ValueError('Order not in orders')
        if format not in ['ase', 'phonopy']:
            raise ValueError('Format must be either ase or phonopy')

        # generate fc array for phonopy format
        fc_array = np.zeros((self.n_atoms, ) * order + (3, ) * order)
        for cluster in product(range(self.n_atoms), repeat=order):
            fc_array[cluster] = self[cluster]

        if format == 'ase':
            if order != 2:
                raise ValueError('ASE format works only for order 2')
            return fc_array.transpose([0, 2, 1, 3]).reshape(
                self.n_atoms * 3, self.n_atoms * 3)
        else:
            return fc_array

    def compute_gamma_frequencies(self) -> np.ndarray:
        """ Returns the Gamma frequencies in THz using the second-order force
        constants. """
        fc2_array = self.get_fc_array(order=2)
        masses = self.supercell.get_masses()
        return _compute_gamma_frequencies(fc2_array, masses)

    def assert_acoustic_sum_rules(self, order: int = None, tol: float = 1e-6):
        """ Asserts that force constants obey acoustic sum rules.

        Parameters
        ----------
        order
            specifies which order to check, if None all are checked
        tol
            numeric tolerance for checking sum rules

        Raises
        ------
        AssertionError
            if acoustic sum rules are violated
        """

        # set up orders
        if order is None:
            orders = self.orders
        else:
            if order not in self.orders:
                raise ValueError('Order not available in FCS')
            orders = [order]

        atomic_indices = range(self.n_atoms)
        for order in orders:
            assert_msg = 'Acoustic sum rule for order {} violated for atom'
            assert_msg += ' {}' * (order - 1) + ' x'
            for ijk in product(atomic_indices, repeat=order-1):
                fc_sum_ijk = np.zeros((3, )*order)
                for m in atomic_indices:
                    cluster = ijk + (m, )
                    fc_sum_ijk += self[cluster]
                assert np.linalg.norm(fc_sum_ijk) < tol, assert_msg.format(order, *ijk)

    def print_force_constant(self, cluster: Tuple[int]) -> None:
        """
        Prints force constants for a cluster in a nice format.

        Parameters
        ----------
        cluster
            sites belonging to the cluster
        """
        print(self._repr_fc(cluster))

    def __eq__(self, other):

        # check supercells are the same
        if not len(self.supercell) == len(other.supercell):
            return False
        if not np.allclose(self.supercell.positions, other.supercell.positions):
            return False
        if not np.allclose(self.supercell.cell, other.supercell.cell):
            return False
        if not all(self.supercell.numbers == other.supercell.numbers):
            return False

        # check orders and clusters are the same
        if self.orders != other.orders:
            return False
        if self.clusters != other.clusters:
            return False

        # check force constants
        for c in self.clusters:
            if not np.allclose(self[c], other[c]):
                return False
        return True

    def __str__(self) -> str:
        s = []
        s.append(' ForceConstants '.center(54, '='))
        s.append('Orders: {}'.format(self.orders))
        s.append('Atoms: {}'.format(self.supercell))
        s.append('')
        if len(self) > 10:
            for c in self.clusters[:3]:
                s.append(self._repr_fc(c)+'\n')
            s.append('...\n')
            for c in self.clusters[-3:]:
                s.append(self._repr_fc(c)+'\n')
        else:
            for c in self.clusters:
                s.append(self._repr_fc(c)+'\n')
        return '\n'.join(s)

    def __repr__(self) -> str:
        fc_dict_str = '{{{}: {}, ..., {}: {}}}'.format(
            self.clusters[0], self[self.clusters[0]].round(5),
            self.clusters[-1], self[self.clusters[-1]].round(5))
        return ('ForceConstants(fc_dict={}, atoms={!r})'
                .format(fc_dict_str, self.supercell))

    def _repr_fc(self, cluster: Tuple[int],
                 precision: float = 5, suppress: bool = True) -> str:
        """
        Representation for single cluster and its force constant.

        Parameters
        ----------
        cluster
            tuple of ints indicating the sites belonging to the cluster
        """
        s = []
        s.append('Cluster: {}'.format(cluster))
        for atom_index in cluster:
            s.append(str(self.supercell[atom_index]))
        s.append('Force constant:')
        s.append(np.array_str(self[cluster], precision=precision,
                 suppress_small=suppress))
        return '\n'.join(s)

    def _sanity_check_dict(self, fc_dict: dict) -> None:
        """ Checks that all indices in clusters are between 0 and number of
        atoms. """
        # Use flat numpy array for fast check
        cluster_indices = np.concatenate([cluster for cluster in fc_dict.keys()])
        if not np.all((0 <= cluster_indices) & (cluster_indices < self.n_atoms)):
            # If the check failed, do slower list processing to find the erring cluster
            for cluster in fc_dict.keys():
                if not all(0 <= i < self.n_atoms for i in cluster):
                    raise ValueError('Cluster {} not in supercell'.format(cluster))

    @classmethod
    def from_arrays(cls, supercell: Atoms,
                    fc2_array: np.ndarray = None, fc3_array: np.ndarray = None):
        """ Constructs FCs from numpy arrays.

        One or both of fc2_array and fc3_array must not be None

        Parameters
        ----------
        supercell
            supercell structure
        fc2_array
            second-order force constant in phonopy format, i.e. must have shape (N, N, 3, 3)
        fc3_array
            third-order force constant in phonopy format, i.e. must have shape (N, N, N, 3, 3, 3)
        """
        if fc2_array is None and fc3_array is None:
            raise ValueError('Please provide force constant arrays.')

        n_atoms = len(supercell)
        if fc2_array is None:
            fc2_dict = dict()
        else:
            if fc2_array.shape != (n_atoms, n_atoms, 3, 3):
                raise ValueError('fc2 array has wrong shape')
            fc2_dict = array_to_dense_dict(fc2_array)

        if fc3_array is None:
            fc3_dict = dict()
        else:
            if fc3_array.shape != (n_atoms, n_atoms, n_atoms, 3, 3, 3):
                raise ValueError('fc2 array has wrong shape')
            fc3_dict = array_to_dense_dict(fc3_array)

        fc_dict = {**fc2_dict, **fc3_dict}
        return cls.from_dense_dict(fc_dict, supercell)

    @classmethod
    def from_sparse_dict(cls, fc_dict: dict, supercell: Atoms):
        """ Assumes label symmetries, meaning only one cluster for each
        permuation should be included

        Parameters
        ----------
        fc_dict
            keys corresponding to clusters and values to the force constants
        supercell
            atomic configuration
        """
        return SortedForceConstants(fc_dict, supercell=supercell)

    @classmethod
    def from_dense_dict(cls, fc_dict: dict, supercell: Atoms):
        """ All permutations of clusters that are not zero must be listed,
        if label symmetries are fullfilled will return a SortedForceConstants

        Parameters
        ----------
        fc_dict
            keys corresponding to clusters and values to the force constants
        supercell
            atomic configuration
        """
        if check_label_symmetries(fc_dict):
            fc_sparse_dict = dense_dict_to_sparse_dict(fc_dict)
            return SortedForceConstants(fc_sparse_dict, supercell=supercell)
        else:
            return RawForceConstants(fc_dict, supercell=supercell)

    @classmethod
    def read_phonopy(cls, supercell: Atoms, fname: str, format: str = None):
        """ Reads force constants from a phonopy calculation.

        Parameters
        ----------
        supercell
            supercell structure (`SPOSCAR`)
        fname
            name of second-order force constant file
        format
            format for second-order force constants;
            possible values: "text", "hdf5"
        """
        fc2_array = phonopyIO.read_phonopy_fc2(fname, format=format)
        return cls.from_arrays(supercell, fc2_array)

    @classmethod
    def read_phono3py(cls, supercell: Atoms, fname: str):
        """ Reads force constants from a phono3py calculation.

        Parameters
        ----------
        supercell
            supercell structure (`SPOSCAR`)
        fname
            name of third-order force constant file
        """
        fc3_array = phonopyIO.read_phonopy_fc3(fname)
        return cls.from_arrays(supercell, fc3_array=fc3_array)

    @classmethod
    def read_shengBTE(cls, supercell: Atoms, fname: str, prim: Atoms, symprec=1e-5):
        """ Reads third order force constants from a shengBTE calculation.

        shengBTE force constants will be mapped onto a supercell.

        Parameters
        ----------
        supercell
            supercell structure
        fname
            name of third-order force constant file
        prim
            primitive configuration (must be equivalent to structure used in
            the shengBTE calculation)
        symprec
            structural symmetry tolerance
        """
        fcs = shengBTEIO.read_shengBTE_fc3(fname, prim, supercell, symprec)
        return fcs

    @classmethod
    def read(cls, fname: str):
        """ Reads ForceConstants from file.

        Parameters
        ----------
        fname
            name of file from which to read
        """
        tar_file = tarfile.open(mode='r', name=fname)
        items = read_items_pickle(tar_file, 'fc_dict')
        fc_dict = items['fc_dict']
        fcs_type = items['fcs_type']
        supercell = read_ase_atoms_from_tarfile(tar_file, 'supercell')
        tar_file.close()
        if fcs_type == 'SortedForceConstants':
            return SortedForceConstants(fc_dict, supercell)
        elif fcs_type == 'RawForceConstants':
            return RawForceConstants(fc_dict, supercell)
        else:
            raise ValueError('FCS type not recongnized')

    def write(self, fname: str) -> None:
        """ Writes entire ForceConstants object to file.

        Parameters
        ----------
        fname
            name of file to which to write
        """
        tar_file = tarfile.open(name=fname, mode='w')
        items_pickle = dict(fc_dict=self._fc_dict, fcs_type=self.__class__.__name__)
        add_items_to_tarfile_pickle(tar_file, items_pickle, 'fc_dict')
        add_ase_atoms_to_tarfile(tar_file, self.supercell, 'supercell')
        tar_file.close()

    def write_to_phonopy(self, fname: str, format: str = None) -> None:
        """
        Writes force constants in phonopy format.

        Parameters
        ----------
        fname
            name of file to which to write second-order force constant
        format
            format for second-order force constants;
             possible values: "text", "hdf5"
        """
        phonopyIO.write_phonopy_fc2(fname, self, format=format)

    def write_to_phono3py(self, fname: str) -> None:
        """
        Writes force constants in phono3py format.

        Parameters
        ----------
        fname
            name of file to which to write third-order force constant
        """
        phonopyIO.write_phonopy_fc3(fname, self)

    def write_to_shengBTE(self, fname: str, prim: Atoms, **kwargs) -> None:
        """
        Writes third order force constants in shengBTE format.

        Parameters
        ----------
        fname
            name of file to which to write third-order force constant
        prim
            primitive configuration (must be equivalent to structure used in
            the shengBTE calculation)
        """
        shengBTEIO.write_shengBTE_fc3(fname, self, prim, **kwargs)


class SortedForceConstants(ForceConstants):
    """ Force constants with label symmetries.

    Parameters
    ----------
    fc_dict : dict
        keys corresponding to clusters and values to the force constants,
        should only contain sorted clusters
    supercell : ase.Atoms
    """

    def __init__(self, fc_dict: dict, supercell: Atoms) -> None:
        super().__init__(supercell)
        self._sanity_check_dict(fc_dict)
        self._fc_dict = fc_dict
        self._orders = sorted(set(len(c) for c in self._fc_dict.keys()))

    def __getitem__(self, cluster: Tuple[int]):
        sorted_cluster = tuple(sorted(cluster))

        # cluster not in fcs
        if sorted_cluster not in self._fc_dict.keys():
            return np.zeros((3,)*len(cluster))

        # return fc for the unsorted cluster
        inv_perm = np.argsort(np.argsort(cluster))
        return self._fc_dict[sorted_cluster].transpose(inv_perm)

    @property
    def orders(self) -> List[int]:
        """ orders for which force constants exist """
        return self._orders.copy()

    def _sanity_check_dict(self, fc_dict: dict) -> None:
        super()._sanity_check_dict(fc_dict)

        # also check clusters are sorted
        for cluster in fc_dict.keys():
            if cluster != tuple(sorted(cluster)):
                raise ValueError('Found unsorted cluster {}'.format(cluster))

    def write_to_GPUMD(self, fname_fc, fname_clusters, order, tol=1e-10):
        """
        Writes force constants of the specified order in GPUMD format.

        Parameters
        ----------
        fname_fc : str
            name of file which contains the lookup force constants
        fname_clusters : str
            name of file which contains the clusters and the fc lookup index
        order : int
            force constants for this order will be written to file
        tol : float
            if the norm of a force constant is less than tol then it is not written.
            if two force-constants are within tol; they are considered equal.
        """
        gpumdIO.write_fcs_gpumd(fname_fc, fname_clusters, self, order, tol)


class RawForceConstants(ForceConstants):
    """ Force constants without label symmetries.

    Parameters
    ----------
    fc_dict : dict
        keys corresponding to clusters and values to the force constants,
        should contain all clusters with nonzero force constants
    supercell : ase.Atoms
    """

    def __init__(self, fc_dict: dict, supercell: Atoms) -> None:
        super().__init__(supercell)
        self._sanity_check_dict(fc_dict)
        self._fc_dict = fc_dict
        self._orders = sorted(set(len(c) for c in self._fc_dict.keys()))

    def __getitem__(self, cluster: Tuple[int]):
        if cluster not in self._fc_dict.keys():
            return np.zeros((3,)*len(cluster))
        return self._fc_dict[cluster]

    @property
    def orders(self) -> List[int]:
        """ orders for which force constants exist """
        return self._orders.copy()


# ====================== #
#     HELPER FUNCTIONS   #
# ====================== #


def array_to_dense_dict(fc_array: np.ndarray, fc_tol: float = 1e-10) -> dict:
    """ Constructs a dense dict from an fc array in phonopy format.

    Force constants with norm smaller than fc_tol will be considered zero and
    therefore not included in the fc_dict.

    Parameters
    ----------
    fc_array
        force constant array in phonopy format
    fc_tol
        tolerance for considering force constants zero or not
    """

    # sanity check
    n_atoms = fc_array.shape[0]
    order = int(len(fc_array.shape) / 2)
    if fc_array.shape != (n_atoms, ) * order + (3, ) * order:
        raise ValueError('fc array has bad shape')

    # construct dense dict
    fc_dict = dict()
    for cluster in product(range(n_atoms), repeat=order):
        fc = fc_array[cluster]
        if np.linalg.norm(fc) > fc_tol:
            fc_dict[cluster] = fc
    return fc_dict


def check_label_symmetries(fc_dict: dict) -> bool:
    """ Checks label symmetries for dense fc dict.

    TODO
    ----
    tol, which one to use etc

    Parameters
    ----------
    fc_dict
        keys corresponding to clusters and values to the force constants
    """
    for cluster, fc in fc_dict.items():
        inv_perm = np.argsort(np.argsort(cluster))
        sorted_cluster = tuple(sorted(cluster))
        if sorted_cluster not in fc_dict:
            return False
        if not np.allclose(fc, fc_dict[sorted_cluster].transpose(inv_perm)):
            return False
    return True


def dense_dict_to_sparse_dict(fc_dict: dict) -> dict:
    """ Converts dense dict to sparse dict.

    This does not check if label symmetry is True, but rather will just keep
    the sorted clusters and their force constants.

    Parameters
    ----------
    fc_dict
        keys corresponding to clusters and values to the force constants
    """
    fc_dict_sparse = dict()
    for cluster, fc in fc_dict.items():
        if cluster == tuple(sorted(cluster)):
            fc_dict_sparse[cluster] = fc
    return fc_dict_sparse


def symbolize_force_constant(fc: np.ndarray, tol: float = 1e-10) -> np.ndarray:
    """Carries out a symbolic symmetrization of a force constant tensor.

    Parameters
    ----------
    fc
        force constant tensor
    tol
        tolerance used to decide whether two elements are identical

    Returns
    -------
    symbolic representation of force constant matrix
    """
    fc_int = np.round(fc / tol).astype(int)
    fc_chars = np.empty(fc_int.shape, dtype=object).flatten()
    all_chars = ascii_lowercase + ascii_uppercase
    lookup_chars = {}
    for i, val in enumerate(fc_int.flatten()):
        if val == 0:
            fc_chars[i] = 0
        elif val in lookup_chars.keys():
            fc_chars[i] = lookup_chars[val]
        elif -val in lookup_chars.keys():
            fc_chars[i] = '-{:}'.format(lookup_chars[-val])
        else:
            lookup_chars[val] = all_chars[len(lookup_chars.keys())]
            fc_chars[i] = lookup_chars[val]
    return fc_chars.reshape(fc_int.shape)


def _compute_gamma_frequencies(fc2: np.ndarray, masses: List[float]) -> np.ndarray:
    """ Computes Gamma frequencies using second-order force constants.
    Assumes fc2 is in units of eV/A2.

    Parameters
    ----------
    fc2
        second-order force constants in phonopy format
    masses
        mass of each atom

    Returns
    -------
    Gamma frequencies in THz
    """

    n_atoms = fc2.shape[0]
    if len(masses) != n_atoms:
        raise ValueError('Length of masses not compatible with fc2')
    mass_matrix = np.sqrt(np.outer(masses, masses))

    # divide with mass matrix
    fc2_tmp = np.zeros((n_atoms, n_atoms, 3, 3))
    for pair in product(range(n_atoms), repeat=2):
        fc2_tmp[pair] = fc2[pair] / mass_matrix[pair]

    # reshape into matrix and solve eigenvalues
    fc2_tmp = fc2_tmp.transpose([0, 2, 1, 3]).reshape(n_atoms * 3, n_atoms * 3)
    eigen_vals, _ = eig(fc2_tmp)
    eigen_vals *= 1e20 / units.J * units.kg  # [1/s**2]
    eigen_vals.sort()

    # if negative eigenval, set frequency to negative
    gamma_frequencies = []
    for val in eigen_vals.real:
        if val >= 0:
            gamma_frequencies.append(np.sqrt(val))
        else:
            gamma_frequencies.append(-np.sqrt(np.abs(val)))

    # Sort and convert to THz
    gamma_frequencies = np.array(gamma_frequencies) / 1e12 / (2 * np.pi)
    gamma_frequencies.sort()
    return gamma_frequencies



from itertools import product
import numpy as np


def _obj2str(a, none_char='-'):
    """ Casts object a to str. """
    if isinstance(a, float):
        # if the float is 2.49999999 then round
        if str(a)[::-1].find('.') > 5:
            return '{:.5f}'.format(a)
    elif a is None:
        return none_char
    return str(a)


_array2str = np.vectorize(_obj2str)


def print_table(matrix: np.ndarray,
                include_sum: bool = False) -> None:
    """ Prints matrix data in a nice table format.

    The matrix element matrix[i][j] should correspond to information about
    order j+2 and n-body i+1.

    Example
    --------
    >> matrix = numpy.array([[None, None], [4.0, 3.0]])
    >> print_table(matrix)

     body/order |  2  |  3
    ------------------------
         1      |  -  |  -
         2      | 4.0 | 3.0

    Parameters
    ----------
    matrix
        matrix to be printed
    include_sum
        whether or not to print the sum along each row and column
    """
    table_str = table_array_to_string(matrix, include_sum)
    print(table_str)


def table_array_to_string(matrix: np.ndarray,
                          include_sum: bool = False) -> str:
    """Generate nice table string from a numpy array with floats/ints.

    Parameters
    ----------
    matrix
        matrix to be printed
    include_sum
        whether or not to print the sum along each row and column
    """
    table_array = _generate_table_array(matrix, include_sum)
    table_array_str = _array2str(table_array)
    table_str = _generate_table_str(table_array_str)
    return table_str


def _generate_table_array(table_array: np.ndarray,
                          include_sum: bool = False):
    """Generate table in numpy array format.

    Parameters
    ----------
    table_array
        matrix to be printed
    include_sum
        whether or not to print the sum along each row and column
    """

    # initialze table
    n_rows, n_cols = table_array.shape
    A = _build_table_frame(order=n_cols+1, nbody=n_rows, include_sum=include_sum)

    # fill table
    for order, nbody in product(range(2, n_cols+2), range(1, n_rows+1)):
        if nbody <= order:
            A[nbody, order-1] = table_array[nbody-1, order-2]

    if include_sum:
        for i, row in enumerate(A[1:-1, 1:-1], start=1):
            A[i, -1] = sum(val for val in row if val is not None)
        for i, col in enumerate(A[1:-1, 1:-1].T, start=1):
            A[-1, i] = sum(val for val in col if val is not None)
        A[-1, -1] = ''

    return A


def _generate_table_str(table_array: np.ndarray) -> str:
    """Generate a string from a numpy array of strings.

    Parameters
    ----------
    table_array
        matrix to be printed
    """
    table_str = []
    n_rows, n_cols = table_array.shape

    # find maximum widths for each column
    widths = []
    for i in range(n_cols):
        widths.append(max(len(val) for val in table_array[:, i])+2)

    # formatting str for each row
    row_format = '|'.join('{:^'+str(width)+'}' for width in widths)

    # finalize
    for i in range(n_rows):
        if i == 1:
            table_str.append('-' * (sum(widths)+n_cols-1))
        table_str.append(row_format.format(*table_array[i, :]))
    table_str = '\n'.join(table_str)
    return table_str


def _build_table_frame(order: int,
                       nbody: int,
                       include_sum: bool = False):
    """ Builds/initializes table/array. """
    if include_sum:
        A = np.empty((nbody+2, order+1), dtype='object')
        A[0, -1] = 'sum'
        A[-1, 0] = 'sum'
    else:
        A = np.empty((nbody+1, order), dtype='object')

    A[0][0] = 'body/order'
    A[0, 1:order] = range(2, order+1)
    A[1:nbody+1, 0] = range(1, nbody+1)
    return A


if __name__ == '__main__':
    # input dummy cutoff table
    # insert row for nbody=1
    cutoffs = np.array([[None, None, None, None, None],
                        [6.0, 6.0, 6.0, 3.7, 3.7],
                        [5.0, 5.0, 5.0, 3.0, 3.0],
                        [3.7, 3.7, 3.7, 0.0, 0.0]])

    # input dummy cluster count table
    cluster_counts = np.array([[1, 3, 5, 5, 2],
                               [12, 22, 39, 42, 58],
                               [19, 41, 123, 421, 912],
                               [42, 112, 410, 617, 3271]])

    print_table(cutoffs)
    print('\n')
    print_table(cluster_counts, include_sum=False)
    print('\n')
    print_table(cluster_counts, include_sum=True)



import numpy as np


def site_offset_to_spos(site, offset, basis_spos):
    """ Returns the scaled position of an atom at specified site and offset
    relative to the basis in scaled coordinates"""
    return offset + basis_spos[site]


def spos_to_pos(spos, cell):
    """ Returns the Cartesian coordinate given the scaled coordinate and cell
    metric (cell vectors as rows)"""
    return np.dot(spos, cell)


def pos_to_spos(pos, cell):
    """ Inverse of sps_to_pos"""
    return np.linalg.solve(cell.T, pos)


def spos_to_site_offset(spos, basis_spos, symprec):
    """ Returns the site and offset of the atom at the specified scaled
    coordinate given the scaled positions of the basis atoms"""
    for site, sp in enumerate(basis_spos):
        offset = spos - sp
        rounded_offset = offset.round(0).astype(np.int64)
        # TODO: fix tolerance (symprec should be w.r.t. cart. coord.)
        if np.allclose(rounded_offset, offset, rtol=0, atol=symprec):
            return site, rounded_offset
    raise Exception('spos {} not compatible with basis {} using symprec {}'
                    .format(spos, basis_spos, symprec))


def pos_to_site_offset(pos, cell, basis_spos, symprec):
    """ helper to map pos -> spos -> site/offset"""
    spos = pos_to_spos(pos, cell)
    return spos_to_site_offset(spos, basis_spos, symprec)


def site_offset_to_pos(site, offset, cell, basis_spos):
    """ helper to map site/offset -> spos -> pos"""
    spos = site_offset_to_spos(site, offset, basis_spos)
    return spos_to_pos(spos, cell)


class BaseAtom:
    """ This class represents an atom placed in an infinite crustal"""
    def __init__(self, site, offset):
        assert type(site) is int, type(site)
        assert len(offset) == 3, len(offset)
        assert (all(type(i) is int for i in offset) or
                all(type(i) is np.int64 for i in offset)), type(offset[0])
        self._site = site
        self._offset = np.array(offset)

    @property
    def site(self):
        return self._site

    @property
    def offset(self):
        return self._offset

    def astype(self, dtype):
        """ Useful arguments: list, tuple, np.int64"""
        return dtype((self._site, *self._offset))


class Atom(BaseAtom):
    """ This class represents a crystal atom in a given structure"""
    def __init__(self, *args, **kwargs):
        self._structure = kwargs.pop('structure', None)
        super().__init__(*args, **kwargs)

    @property
    def pos(self):
        return site_offset_to_pos(self._site, self._offset,
                                  self._structure.cell,
                                  self._structure.spos)

    @property
    def number(self):
        return self._structure.numbers[self._site]


class SupercellAtom(Atom):
    """ Represents an atom in a supercell but site and offset given by an
    underlying primitve cell"""
    def __init__(self, *args, **kwargs):
        self._index = kwargs.pop('index')
        assert type(self._index) is int
        super().__init__(*args, **kwargs)

    @property
    def index(self):
        return self._index


class Structure:
    """ This class essentially wraps the ase.Atoms class but is a bit more
    carefull about pbc and scaled coordinates. It also returns hiphive.Atom
    objects instead"""
    def __init__(self, atoms, symprec=1e-6):
        spos = atoms.get_scaled_positions(wrap=False)
        for sp in spos.flat:
            if not (-symprec < sp < (1 - symprec)):
                raise ValueError('bad spos {}'.format(sp))
        self._spos = spos
        self._cell = atoms.cell
        self._numbers = atoms.numbers

    def __len__(self):
        return len(self._spos)

    @property
    def spos(self):
        return self._spos

    @property
    def cell(self):
        return self._cell

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError('Structure contains {} atoms'.format(len(self)))
        return Atom(index, (0, 0, 0), structure=self)

    def atom_from_pos(self, pos, symprec=None):
        if symprec is None:
            symprec = self._symprec
        site, offset = pos_to_site_offset(pos, self._cell, self._spos, symprec)
        return Atom(site, offset, structure=self)


class Supercell:
    """ This class tries to represent atoms in a supercell as positioned on the
    primitve lattice"""
    def __init__(self, supercell, prim, symprec):
        self._supercell = Structure(supercell)
        self._prim = Structure(prim)
        self._symprec = symprec
        self._map = list()
        self._inverse_map_lookup = dict()
        self._create_map()

    def _create_map(self):
        for atom in self._supercell:
            atom = self._prim.atom_from_pos(atom.pos, self._symprec)
            self._map.append(atom.astype(tuple))

    def wrap_atom(self, atom):
        atom = Atom(atom.site, atom.offset, structure=self._prim)
        tup = atom.astype(tuple)
        index = self._inverse_map_lookup.get(tup, None)
        if index is None:
            atom = self._supercell.atom_from_pos(atom.pos, self._symprec)
            index = atom.site
            self._inverse_map_lookup[tup] = index
        return self[index]

    def index(self, site, offset):
        atom = self.wrap_atom(BaseAtom(site, offset))
        return atom.index

    def __getitem__(self, index):
        tup = self._map[index]
        return SupercellAtom(tup[0], tup[1:], structure=self._prim,
                             index=index)

    def __len__(self):
        return len(self._supercell)



import pickle
import numpy as np
from ase.neighborlist import NeighborList
from hiphive.input_output.pretty_table_prints import table_array_to_string


class Cutoffs:
    """ This class maintains information about the cutoff configuration,
    i.e. which clusters will be included (="inside cutoff"). It also
    encapsulates functionality that is used e.g., during cluster space
    construction.

    Here, `n-body` refers to number of atoms in a cluster. For example
    the cluster (0011) is a two-body cluster of fourth order and the
    cluster (123) is a three-body cluster of third order.

    Parameters
    ----------
    cutoff_matrix : numpy.ndarray
        the matrix element `ij` provides to the cutoff for order `j+2`
        and nbody `i+2`; elements with `i>j` will be ignored
    """

    def __init__(self, cutoff_matrix):

        self._cutoff_matrix = np.array(cutoff_matrix, dtype=float)

        if len(self._cutoff_matrix.shape) != 2:
            raise ValueError('Please specify cutoff matrix as a 2D array')
        for i, row in enumerate(self._cutoff_matrix):
            if np.any(row[i:] < 0):
                raise ValueError('Negative number as cutoff')
            row[:i] = np.nan
        self._cutoff_matrix = self._cutoff_matrix[:(self.max_nbody-1), :(self.max_order-1)]

    @property
    def cutoff_matrix(self):
        """ numpy.ndarray : copy of cutoff matrix """
        return self._cutoff_matrix.copy()

    @property
    def orders(self):
        """ list(int) : allowed orders """
        return list(range(2, self.max_order + 1))

    @property
    def nbodies(self):
        """ list(int) : allowed bodies """
        return list(range(2, self.max_nbody + 1))

    def get_cutoff(self, order, nbody):
        """
        Returns cutoff for a given body and order.

        Parameters
        ----------
        order : int
        nbody : int

        Raises
        ------
        ValueError
            if order is not in orders
        ValueError
            if nbody is not in nbodies
        ValueError
            if nbody is larger than order

        Returns
        -------
        float
        """
        if order not in self.orders:
            raise ValueError('order not in orders')
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        if nbody > order:
            raise ValueError('nbody can not be larger than order')
        return self._cutoff_matrix[nbody - 2, order - 2]

    @property
    def max_cutoff(self):
        """ float : maximum cutoff """
        max_cutoff = 0
        for i, row in enumerate(self._cutoff_matrix):
            max_cutoff = max(max_cutoff, np.max(row[i:]))
        return max_cutoff

    @property
    def max_nbody(self):
        """ int : maximum body """
        nbody = 1
        for i, row in enumerate(self._cutoff_matrix):
            if np.any(row[i:]):
                nbody = i + 2
        return nbody

    @property
    def max_order(self):
        """ int : maximum order """
        order = None
        for col in range(self._cutoff_matrix.shape[1]):
            if np.any(self._cutoff_matrix[:col + 1, col]):
                order = col + 2
        return order

    def max_nbody_cutoff(self, nbody):
        """ Return maximum cutoff for a given body. """
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        return np.max(self._cutoff_matrix[nbody - 2, max(0, nbody - 2):])

    def max_nbody_order(self, nbody):
        """ Returns maximum order for a given body """
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        row = self._cutoff_matrix[nbody - 2]
        max_order = None
        for order, cutoff in enumerate(row[nbody-2:], start=nbody):
            if cutoff:
                max_order = order
        return max_order

    def write(self, fileobj):
        """ Writes instance to file.

        Parameters
        ----------
        fileobj : file-like object
            file-like object to which the cutoffs will be written to
        """
        pickle.dump(self._cutoff_matrix, fileobj)

    def read(fileobj):
        """ Reads an instance from file.

        Parameters
        ----------
        fileobj : file-like object
            input file to read from
        """
        data = pickle.load(fileobj)
        return Cutoffs(data)

    def to_filename_tag(self):
        """ Simple function turning cutoffs into a string to be used in e.g.
        filenames. """
        s = []
        for i, c in enumerate(self._cutoff_matrix.tolist(), start=2):
            s.append('{}body-{}'.format(i, '_'.join(map(str, c))))
        return '_'.join(s)

    def __str__(self):
        cutoff_matrix = self._cutoff_matrix.copy()
        cutoff_matrix = np.vstack(([[None] * len(self.orders)], cutoff_matrix))
        s = table_array_to_string(cutoff_matrix)

        width = max(len(c) for c in s.split('\n'))
        header = ' Cutoffs '.center(width, '=') + '\n'
        bottom = '\n' + ''.center(width, '=')
        s = header + s + bottom
        return s

    def __repr__(self):
        return 'Cutoffs({!r})'.format(self._cutoff_matrix)


class CutoffMaximumBody(Cutoffs):
    """ Specify cutoff-list plus maximum body

    Usefull when creating e.g. 6th order expansions but with only 3-body
    interactions.

    Parameters
    ----------
    cutoff_list : list
        list of cutoffs for order 2, 3, etc. Must be in decresing order
    max_nbody : int
        No clusters containing more than max_nbody atoms will be generated
    """

    def __init__(self, cutoff_list, max_nbody):
        cutoff_matrix = np.zeros((max_nbody - 1, len(cutoff_list)))
        for order, cutoff in enumerate(cutoff_list, start=2):
            cutoff_matrix[:, order - 2] = cutoff
        super().__init__(cutoff_matrix)


def is_cutoff_allowed(atoms, cutoff):
    """ Checks if atoms is compatible with cutoff

    Parameters
    ----------
    atoms : ase.Atoms
        structure used for checking compatibility with cutoff
    cutoff : float
        cutoff to be tested

    Returns
    -------
    bool
        True if cutoff compatible with atoms object, else False
    """
    nbrlist = NeighborList(cutoffs=[cutoff / 2] * len(atoms), skin=0,
                           self_interaction=False, bothways=True)
    nbrlist.update(atoms)

    for i in range(len(atoms)):
        neighbors, _ = nbrlist.get_neighbors(i)
        if i in neighbors:
            return False
        if len(neighbors) != len(set(neighbors)):
            return False
    return True


def estimate_maximum_cutoff(atoms, max_iter=11):
    """ Estimates the maximum possible cutoff given the atoms object

    Parameters
    ----------
    atoms : ase.Atoms
        structure used for checking compatibility with cutoff
    max_iter : int
        number of iterations in binary search
    """

    # First upper boundary of cutoff
    upper_cutoff = min(np.linalg.norm(atoms.cell, axis=1))

    # generate all possible offsets given upper_cutoff
    nbrlist = NeighborList(cutoffs=[upper_cutoff / 2] * len(atoms), skin=0,
                           self_interaction=False, bothways=True)
    nbrlist.update(atoms)
    all_offsets = []
    for i in range(len(atoms)):
        _, offsets = nbrlist.get_neighbors(i)
        all_offsets.extend([tuple(offset) for offset in offsets])

    # find lower boundary and new upper boundary
    unique_offsets = set(all_offsets)
    unique_offsets.discard((0, 0, 0))
    upper = min(np.linalg.norm(np.dot(offset, atoms.cell))
                for offset in unique_offsets)
    lower = upper / 2.0

    # run binary search between the upper and lower bounds
    for _ in range(max_iter):
        cutoff = (upper + lower) / 2
        if is_cutoff_allowed(atoms, cutoff):
            lower = cutoff
        else:
            upper = cutoff
    return lower


class BaseClusterFilter:
    """Base cluster filter class.

    This filter simply accepts all proposed clusters.  A proper
    subclass must implement the same methods.
    """
    def setup(self, atoms):
        """ The filter is passed the environment of the primitive cell.

        Parameters
        ----------
        atoms : ase.Atoms
            non-pbc primitive cell plus neighboring atoms
        """
        self._atoms = atoms

    def __call__(self, cluster):
        """ Returns True or False when a cluster is proposed.

        Parameters
        ----------
        cluster : tuple(int)
            indices of proposed cluster referenced to the internal
            atoms object
        """
        return True



#!/usr/bin/env python3

import re
import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 8, 0, 'final', 0):
    raise SystemExit('Python 3.8 or later is required!')

with open('README.rst', encoding='utf-8') as fd:
    long_description = fd.read()

with open('hiphive/__init__.py') as fd:
    lines = '\n'.join(fd.readlines())

version = re.search("__version__ = '(.*)'", lines).group(1)
maintainer = re.search("__maintainer__ = '(.*)'", lines).group(1)
maintainer_email = re.search("__maintainer_email__ = '(.*)'", lines).group(1)
url = re.search("__url__ = '(.*)'", lines).group(1)
license = re.search("__license__ = '(.*)'", lines).group(1)
description = re.search("__description__ = '(.*)'", lines).group(1)

# PyPI name
name = 'hiphive'

setup(name=name,
      version=version,
      description=description,
      long_description=long_description,
      url=url,
      maintainer=maintainer,
      maintainer_email=maintainer_email,
      platforms=['unix'],
      install_requires=['ase',
                        'h5py',
                        'numba>=0.55',
                        'numpy>=1.18',  # imposed by numba
                        'pandas',
                        'scipy>=1.0.0',  # imposed by numba
                        'scikit-learn',
                        'spglib>=2.5',
                        'sympy>=1.1',
                        'trainstation>=0.2'],
      packages=find_packages(),
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
          'Intended Audience :: Science/Research',
          f'License :: OSI Approved :: {license}',
          'Topic :: Scientific/Engineering :: Physics'])



import numpy as np
from hiphive.force_constant_model import ForceConstantModel
from hiphive import StructureContainer
from hiphive.utilities import prepare_structures
from hiphive.structure_generation import generate_rattled_structures, \
                                         generate_phonon_rattled_structures
from trainstation import Optimizer


def self_consistent_harmonic_model(atoms_ideal, calc, cs, T, alpha,
                                   n_iterations, n_structures,
                                   QM_statistics=False, parameters_start=None,
                                   imag_freq_factor=1.0, fit_kwargs={}):
    """
    Constructs a set of self-consistent second-order force constants
    that provides the closest match to the potential energy surface at
    a the specified temperature.

    Parameters
    ----------
    atoms_ideal : ase.Atoms
        ideal structure
    calc : ASE calculator object
        `calculator
        <https://wiki.fysik.dtu.dk/ase/ase/calculators/calculators.html>`_
        to be used as reference potential
    cs : ClusterSpace
        clusterspace onto which to project the reference potential
    T : float
        temperature in K
    alpha : float
        stepsize in optimization algorithm
    n_iterations : int
        number of iterations in poor mans
    n_structures : int
        number of structures to use when fitting
    QM_statistics : bool
        if the amplitude of the quantum harmonic oscillator shoule be used
        instead of the classical amplitude in the phonon rattler
    parameters_start : numpy.ndarray
        parameters from which to start the optimization
    image_freq_factor: float
        If a squared frequency, w2, is negative then it is set to
        w2 = imag_freq_factor * np.abs(w2)
    fit_kwargs : dict
        kwargs to be used in the fitting process (via Optimizer)

    Returns
    -------
    list(numpy.ndarray)
        sequence of parameter vectors generated while iterating to
        self-consistency
    """

    if not 0 < alpha <= 1:
        raise ValueError('alpha must be between 0.0 and 1.0')

    if max(cs.cutoffs.orders) != 2:
        raise ValueError('ClusterSpace must be second order')

    # initialize things
    sc = StructureContainer(cs)
    fcm = ForceConstantModel(atoms_ideal, cs)

    # generate initial model
    if parameters_start is None:
        print('Creating initial model')
        rattled_structures = generate_rattled_structures(atoms_ideal, n_structures, 0.03)
        rattled_structures = prepare_structures(rattled_structures, atoms_ideal, calc, False)
        for structure in rattled_structures:
            sc.add_structure(structure)
        opt = Optimizer(sc.get_fit_data(), train_size=1.0, **fit_kwargs)
        opt.train()
        parameters_start = opt.parameters
        sc.delete_all_structures()

    # run poor mans self consistent
    parameters_old = parameters_start.copy()
    parameters_traj = [parameters_old]

    for i in range(n_iterations):
        # generate structures with old model
        print('Iteration {}'.format(i))
        fcm.parameters = parameters_old
        fc2 = fcm.get_force_constants().get_fc_array(order=2, format='ase')
        phonon_rattled_structures = generate_phonon_rattled_structures(
            atoms_ideal, fc2, n_structures, T, QM_statistics=QM_statistics,
            imag_freq_factor=imag_freq_factor)
        phonon_rattled_structures = prepare_structures(
            phonon_rattled_structures, atoms_ideal, calc, False)

        # fit new model
        for structure in phonon_rattled_structures:
            sc.add_structure(structure)
        opt = Optimizer(sc.get_fit_data(), train_size=1.0, **fit_kwargs)
        opt.train()
        sc.delete_all_structures()

        # update parameters
        parameters_new = alpha * opt.parameters + (1-alpha) * parameters_old

        # print iteration summary
        disps = [atoms.get_array('displacements') for atoms in phonon_rattled_structures]
        disp_ave = np.mean(np.abs(disps))
        disp_max = np.max(np.abs(disps))
        x_new_norm = np.linalg.norm(parameters_new)
        delta_x_norm = np.linalg.norm(parameters_old-parameters_new)
        print('    |x_new| = {:.5f}, |delta x| = {:.8f}, disp_ave = {:.5f}, disp_max = {:.5f}, '
              'rmse = {:.5f}'.format(x_new_norm, delta_x_norm, disp_ave, disp_max, opt.rmse_train))
        parameters_traj.append(parameters_new)
        parameters_old = parameters_new

    return parameters_traj



# This module contains the various variables used as tolerances and similar
# throughout the code in the core functionality.


class config:
    class eigensymmetries:
        rotation_integer_tolerance = 1e-12
        method = 'iterative'
        crystal_symmetries = True

        class iterative:
            zero_tolerance = 1e-12
            simplify = True
            simplify_tolerance = None
            method = 'symbolic'


"""This is the tolerance used when the rotation matrices and similar are tested
if they are integers.
"""
integer_tolerance = 1e-12

""" For each symmetry, the constraint matrix can be reduced to square
again. This can be done either by 'symbolic', 'numeric' or not at all
(None). Default is None since the matrix is often small enough to fit
in memory.
'symbolic', 'numeric', None
"""
eigentensor_compress_mode = None

"""If this is True, before every symbolic compression the values will be
simplified by sympy, potentially turning them into exact rational or
irrational numbers. This can be useful for systems with non-integer
rotation matrices in cartesian coordinates e.g. hcp. The main purpose is
to make the rref more stable against repeating rounding errors.
True, False
"""
eigentensor_simplify_before_compress = False

""" If non compress was used during construction but used before solving
True, False
"""
eigentensor_simplify_before_last_compress = False

"""If the compress_mode is None the constraint matrix might be compressed
right before the nullspace() solver
'numeric', 'symbolic', None
"""
eigentensor_compress_before_solve = None

"""This might make the nullspace() more stable
True, False
"""
eigentensor_simplify_before_solve = True

"""
'symbolic', 'numeric'
"""
eigentensor_solve_mode = 'symbolic'


"""
'symbolic', 'numeric'
"""
sum_rule_constraint_mode = 'symbolic'

"""
True, False
"""
sum_rule_constraint_simplify = True


"""
True, False
"""
constraint_vectors_simplify_before_compress = True

"""
'symbolic', 'numeric', None
"""
constraint_vectors_compress_mode = 'symbolic'

"""
True, False
"""
constraint_vectors_simplify_before_solve = True

"""
'symbolic', 'numeric'
"""
constraint_vectors_solve_mode = 'symbolic'



import numpy as np


def compute_sed(traj, ideal, prim, k_points):
    """ Computes spectral energy density for a trajectory.

    Parameters
    ----------
    traj : list
        trajectory with atoms objects with velocities
    ideal : ASE atoms object
        ideal atoms object
    prim : ASE atoms object
        compatible primitive cell. Must be aligned correctly
    k_points : list
        list of k points in cart coord (2pi must be included)
    """

    velocities = []
    for atom in traj:
        velocities.append(atom.get_velocities())
    velocities = np.array(velocities)
    velocities = velocities.transpose(1, 2, 0).copy()
    velocities = np.fft.fft(velocities, axis=2)

    masses = prim.get_masses()
    indices, offsets = _index_offset(ideal, prim)

    pos = np.dot(k_points, np.dot(offsets, prim.cell).T)
    exppos = np.exp(1.0j * pos)
    density = np.zeros((len(k_points), velocities.shape[2]))
    for alpha in range(3):
        for b in range(len(masses)):
            tmp = np.zeros(density.shape, dtype=np.complex)
            for i in range(len(indices)):
                index = indices[i]
                if index != b:
                    continue
                tmp += np.outer(exppos[:, i], velocities[i, alpha])

            density += masses[b] * np.abs(tmp)**2

    return density


def _index_offset(atoms, prim, atol=1e-3, rtol=0.0):
    index, offset = [], []
    for pos in atoms.positions:
        spos = np.linalg.solve(prim.cell.T, pos)
        for i, spos2 in enumerate(prim.get_scaled_positions()):
            off = spos - spos2
            off_round = np.round(off)
            if not np.allclose(off, off_round, atol=atol, rtol=rtol):
                continue
            index.append(i)
            off = off_round.astype(int)
            assert np.allclose(off, off_round)
            offset.append(off)
            break
        else:
            raise ValueError('prim not compatible with atoms')

    index, offset = np.array(index), np.array(offset)
    return index, offset
